#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# This file is part of the PyNCulture project, which aims at providing tools to
# easily generate complex neuronal cultures.
# Copyright (C) 2017 SENeC Initiative
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Importing shapes from files """

import logging

import numpy as np

from shapely.affinity import affine_transform
from shapely.geometry import MultiPolygon
from shapely.ops import cascaded_union

from .shape import Shape
from .tools import pop_largest

try:
    from .units import _unit_support
except ImportError:
    _unit_support = False


# --------------------- #
# Check SVG/DXF support #
# --------------------- #

_logger = logging.getLogger(__name__)
from .pync_log import _log_message

_svg_support = False
_dxf_support = False

try:
    from . import svgtools
    _svg_support = True
except ImportError as e:
    _log_message(
        _logger, "INFO", "SVG import disabled: {}\n".format(e) +\
                         "Install 'svg.path' to use it.")

try:
    from . import dxftools
    _dxf_support = True
except ImportError as e:
    _log_message(_logger, "INFO", "DFX import disabled: {}\n".format(e) +\
                                  "Install 'dxfgrabber' to use it.")


# -------------- #
# Load from file #
# -------------- #

def shapes_from_file(filename, min_x=None, max_x=None, unit='um',
                     parent=None, interpolate_curve=50,
                     default_properties=None, **kwargs):
    '''
    Generate a set of :class:`Shape` objects from an SVG, a DXF, or a WKT/WKB
    file.

    Valid file needs to contain only closed objects among:
    rectangles, circles, ellipses, polygons, and closed curves.
    The objects do not have to be simply connected.

    .. versionadded:: 0.3

    Parameters
    ----------
    filename : str
        Path to the SVG, DXF, or WKT/WKB file.
    min_x : float, optional (default: -5000.)
        Position of the leftmost coordinate of the shape's exterior, in `unit`.
    max_x : float, optional (default: 5000.)
        Position of the rightmost coordinate of the shape's exterior, in
        `unit`.
    unit : str, optional (default: 'um')
        Unit of the positions, among micrometers ('um'), milimeters ('mm'),
        centimeters ('cm'), decimeters ('dm'), or meters ('m').
    parent : :class:`nngt.Graph` or subclass, optional (default: None)
        Assign a parent graph if working with NNGT.
    interpolate_curve : int, optional (default: 50)
        Number of points by which a curve should be interpolated into segments.

    Returns
    -------
    culture : :class:`Shape` object
        Shape, vertically centred around zero, such that
        :math:`min(y) + max(y) = 0`.
    '''
    polygons, points = None, None

    if _unit_support:
        from .units import Q_
        if isinstance(min_x, Q_):
            min_x = min_x.m_as(unit)
        if isinstance(max_x, Q_):
            max_x = max_x.m_as(unit)

    if filename.endswith(".svg") and _svg_support:
        polygons, points = svgtools.polygons_from_svg(
            filename, parent=parent, interpolate_curve=interpolate_curve,
            return_points=True)
    elif filename.endswith(".dxf") and _dxf_support:
        polygons, points = dxftools.polygons_from_dxf(
            filename, parent=parent, interpolate_curve=interpolate_curve,
            return_points=True)
    elif filename.endswith(".wkt"):
        from shapely.wkt import loads
        content = ""
        with open(filename, 'r') as f:
            content = "".join([l for l in f])
        polygons = [loads(content)]
        points = {'path': [np.array(polygons[0].exterior.coords)]}
    elif filename.endswith(".wkb"):
        from shapely.wkb import loads
        content = ""
        with open(filename, 'r') as f:
            content = "".join([l for l in f])
        polygons = [loads(content)]
        points = {'path': [np.array(polygons[0].exterior.coords)]}
    else:
        raise ImportError("You do not have support to load '" + filename + \
                          "', please install either 'shapely', 'svg.path' or "
                          "'dxfgrabber' to enable it.")

    min_x_val = np.inf
    max_x_val = -np.inf
    min_y_val = np.inf
    max_y_val = -np.inf

    # find smallest and highest x values
    for p in polygons:
        min_x_tmp, min_y_tmp, max_x_tmp, max_y_tmp = p.bounds
        if min_x_tmp < min_x_val:
            min_x_val = min_x_tmp
        if max_x_tmp > max_x_val:
            max_x_val = max_x_tmp
        if min_y_tmp < min_y_val:
            min_y_val = min_y_tmp
        if max_y_tmp > max_y_val:
            max_y_val = max_y_tmp

    # set optional shifts if center will change
    y_center     = 0.5*(max_y_val + min_y_val)
    x_shift      = 0
    scale_factor = 1
    if None not in (min_x, max_x):
        scale_factor = (max_x - min_x) / (max_x_val - min_x_val)
        x_shift     += max_x - max_x_val * scale_factor
        y_center    *= scale_factor
    elif min_x is not None:
        x_shift += min_x - min_x_val
    elif max_x is not None:
        x_shift += max_x - max_x_val

    shapes = []

    # scale and shift the shapes
    for p in polygons:
        # define affine transformation (xx, xy, yx, yy, xoffset, yoffset)
        aff_trans = [scale_factor, 0, 0, scale_factor, x_shift, -y_center]
        p_new     = affine_transform(p.buffer(0), aff_trans)
        # check incorrect shapes
        if not np.isclose(p_new.area, p.area*scale_factor**2):
            tolerance = p.length*1e-6
            p_new = affine_transform(p.simplify(tolerance), aff_trans)
            if not np.isclose(p_new.area, p.area*scale_factor**2, 1e-5):
                raise RuntimeError("Error when generating the shape, check "
                                   "your file...")
        x_min, _, x_max, _ = p_new.bounds
        shapes.append(
            Shape.from_polygon(p_new, min_x=x_min, max_x=x_max, unit=unit))

    if kwargs.get("return_points", False):
        return shapes, points
    return shapes


def culture_from_file(filename, min_x=None, max_x=None, unit='um',
                      parent=None, interpolate_curve=50,
                      internal_shapes_as="holes",
                      default_properties=None,
                      other_properties=None):
    '''
    Generate a culture from an SVG, a DXF, or a WKT/WKB file.

    Valid file needs to contain only closed objects among:
    rectangles, circles, ellipses, polygons, and closed curves.
    The objects do not have to be simply connected.

    .. versionchanged:: 0.6
        Added `internal_shapes_as` and `other_properties` keyword parameters.

    Parameters
    ----------
    filename : str
        Path to the SVG, DXF, or WKT/WKB file.
    min_x : float, optional (default: -5000.)
        Position of the leftmost coordinate of the shape's exterior, in `unit`.
    max_x : float, optional (default: 5000.)
        Position of the rightmost coordinate of the shape's exterior, in
        `unit`.
    unit : str, optional (default: 'um')
        Unit of the positions, among micrometers ('um'), milimeters ('mm'),
        centimeters ('cm'), decimeters ('dm'), or meters ('m').
    parent : :class:`nngt.Graph` or subclass, optional (default: None)
        Assign a parent graph if working with NNGT.
    interpolate_curve : int, optional (default: 50)
        Number of points by which a curve should be interpolated into segments.
    internal_shapes_as : str, optional (default: "holes")
        Defines how additional shapes contained in the main environment should
        be processed. If "holes", then these shapes are substracted from the
        main environment; if "areas", they are considered as areas.
    default_properties : dict, optional (default: None)
        Properties of the default area of the culture.
    other_properties : dict, optional (default: None)
        Properties of the non-default areas of the culture (internal shapes if
        `internal_shapes_as` is set to "areas").

    Returns
    -------
    culture : :class:`Shape` object
        Shape, vertically centred around zero, such that
        :math:`min(y) + max(y) = 0`.
    '''
    shapes = shapes_from_file(
        filename, min_x=min_x, max_x=max_x, unit=unit, parent=parent,
        interpolate_curve=interpolate_curve,
        default_properties=default_properties)

    # make sure that the main container contains all other polygons
    main_container = pop_largest(shapes)
    interiors      = [item.coords for item in main_container.interiors]
    invalid_shapes = []

    internal_shapes = []
    for i, s in enumerate(shapes):
        valid = main_container.contains(s)
        if valid:
            internal_shapes.append(s)
        else:
            # because of interpolation, some shapes can go slightly out of
            # the main container, we correct this by subtracting them from the
            # main container afterwards
            valid = s.difference(main_container).area < 1e-2*s.area
            internal_shapes.append(s.intersection(main_container))
        assert valid, "Some polygons are not contained in the main container."

    internal_shapes = \
        cascaded_union(internal_shapes) if internal_shapes else Shape([])

    if internal_shapes_as == "holes":
        diff           = main_container.difference(internal_shapes)
        main_container = Shape.from_polygon(diff, min_x=None, max_x=None)
        interiors      = [item.coords for item in main_container.interiors]
    elif internal_shapes_as != "areas":
        raise ValueError("Invalid value {} for `internal_shapes_as`".format(
                         internal_shapes_as))

    culture  = Shape(main_container.exterior.coords, interiors)
    old_area = culture.area
    # make sure it is a valid Polygon
    culture = Shape.from_polygon(culture.buffer(0), min_x=None, max_x=None,
                                 unit=unit, parent=parent,
                                 default_properties=default_properties)

    if internal_shapes_as == "areas" and not internal_shapes.is_empty:
        if isinstance(internal_shapes, MultiPolygon):
            for i, p in enumerate(internal_shapes):
                culture.add_area(p, name="area_{}".format(i),
                                 properties=other_properties)
        else:
            culture.add_area(internal_shapes, name="area_1",
                             properties=other_properties)

    # check
    if not np.isclose(culture.area, old_area):
        tolerance = culture.length*1e-6
        culture = Shape.from_polygon(culture.simplify(tolerance), min_x=None,
                                     max_x=None, unit=unit, parent=parent,
                                     default_properties=default_properties)
        if not np.isclose(culture.area, old_area, 1e-5):
            raise RuntimeError("Error when generating the culture, check "
                               "your file...")
    return culture



#-*- coding:utf-8 -*-
#
# core/gt_graph.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Graph-tool Graph subclassing """

from collections import OrderedDict
from copy import deepcopy
import logging

import numpy as np
import scipy.sparse as ssp

import nngt
from nngt.lib import InvalidArgument, BWEIGHT, nonstring_container, is_integer
from nngt.lib.connect_tools import (_cleanup_edges, _set_dist_new_edges,
                                    _set_default_edge_attributes)
from nngt.lib.converters import _to_np_array
from nngt.lib.graph_helpers import (_get_dtype, _get_gt_weights,
                                    _post_del_update)
from nngt.lib.logger import _log_message
from .graph_interface import GraphInterface, BaseProperty


logger = logging.getLogger(__name__)


# ---------- #
# Properties #
# ---------- #

class _GtNProperty(BaseProperty):

    ''' Class for generic interactions with nodes properties (graph-tool)  '''

    def __getitem__(self, name):
        dtype = super(_GtNProperty, self).__getitem__(name)

        g = self.parent()._graph

        if dtype == "string":
            return g.vertex_properties[name].get_2d_array([0])[0]
        elif dtype == "object":
            vprop = g.vertex_properties[name]
            return _to_np_array(
                [vprop[i] for i in range(g.num_vertices())], dtype)

        return _to_np_array(g.vertex_properties[name].a, dtype)

    def __setitem__(self, name, value):
        dtype = super(_GtNProperty, self).__getitem__(name)

        g = self.parent()._graph

        if name in self:
            size = g.num_vertices()
            if len(value) == size:
                if dtype == "string":
                    g.vertex_properties[name].set_2d_array(np.asarray(value))
                else:
                    g.vertex_properties[name].a = np.asarray(value)
            else:
                raise ValueError("A list or a np.array with one entry per "
                                 "node in the graph is required")
        else:
            raise InvalidArgument("Attribute does not exist yet, use "
                                  "set_attribute to create it.")

    def new_attribute(self, name, value_type, values=None, val=None):
        dtype = object
        g     = self.parent()._graph

        if val is None:
            if value_type == "int":
                val = int(0)
                dtype = int
            elif value_type == "double":
                val = np.NaN
                dtype = float
            elif value_type == "string":
                val = ""
            else:
                val = None

        if values is None:
            values = _to_np_array([deepcopy(val)
                                   for _ in range(g.num_vertices())], dtype)

        if len(values) != g.num_vertices():
            raise ValueError("A list or a np.array with one entry per "
                             "node in the graph is required, got "
                             "{} values vs {} nodes for '{}'.".format(
                                len(values), g.num_vertices(), name))

        # store name and value type in the dict
        super(_GtNProperty, self).__setitem__(name, value_type)

        # store the real values in the attribute
        nprop = g.new_vertex_property(value_type, vals=values)
        g.vertex_properties[name] = nprop
        self._num_values_set[name] = len(values)

    def set_attribute(self, name, values, nodes=None):
        '''
        Set the node attribute.

        Parameters
        ----------
        name : str
            Name of the node attribute.
        values : array, size N
            Values that should be set.
        nodes : array-like, optional (default: all nodes)
            Nodes for which the value of the property should be set. If `nodes`
            is not None, it must be an array of size N.
        '''
        g = self.parent()._graph

        num_nodes = g.num_vertices()
        num_n = len(nodes) if nodes is not None else num_nodes

        if num_n == num_nodes:
            self[name] = values
            self._num_values_set[name] = num_nodes
        elif num_n:
            if num_n != len(values):
                raise ValueError("`nodes` and `values` must have the same "
                                 "size; got respectively " + str(num_n) + \
                                 " and " + str(len(values)) + " entries.")

            non_obj = (super(_GtNProperty, self).__getitem__(name)
                       not in ('string', 'object'))

            if self._num_values_set[name] == num_nodes - num_n and non_obj:
                g.vertex_properties[name].a[-num_n:] = values
            else:
                for n, val in zip(nodes, values):
                    g.vertex_properties[name][n] = val
        if num_n:
            self._num_values_set[name] = num_nodes


class _GtEProperty(BaseProperty):

    ''' Class for generic interactions with nodes properties (graph-tool)  '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._edges_deleted = False

    def __getitem__(self, name):
        '''
        Return the attributes of an edge or a list of edges.
        '''
        g = self.parent()._graph

        Edge = g.edge

        if isinstance(name, slice):
            eprop = {}
            for k in self:
                ep = g.edge_properties[k]

                if self._edges_deleted:
                    Edge = g.edge
                    data = [ep[Edge(*e)] for e in self.parent().edges_array]

                    eprop[k] = data[name]
                else:
                    eprop[k] = ep.a[name]

            return eprop
        elif nonstring_container(name):
            # name is and edge or contains edges
            if len(name) == 0:
                return {k: [] for k in self}

            eprop = {}
            if nonstring_container(name[0]):
                # name is a list of edges
                for k in self.keys():
                    tmp = g.edge_properties[k]

                    dtype = super().__getitem__(k)

                    eprop[k] = _to_np_array(
                        [tmp[Edge(*e)] for e in name], dtype)
            else:
                # name is a single edge
                for k in self.keys():
                    eprop[k] = g.edge_properties[k][Edge(*name)]

            return eprop

        # name is a string
        dtype = super().__getitem__(name)

        if self._edges_deleted:
            eprop = g.edge_properties[name]
            edges = self.parent().edges_array

            return _to_np_array([eprop[Edge(*e)] for e in edges], dtype)

        if dtype == "string":
            return g.edge_properties[name].get_2d_array([0])[0]
        elif dtype == "object":
            tmp = g.edge_properties[name]

            return _to_np_array([tmp[Edge(*e)] for e in g.get_edges()], dtype)

        return _to_np_array(g.edge_properties[name].a, dtype)

    def __setitem__(self, name, value):
        g = self.parent()._graph

        if name in self:
            size = g.num_edges()

            if size:
                dtype = super(_GtEProperty, self).__getitem__(name)

                # check for list value for one edge
                if dtype == "object" and size == 1 and isinstance(value, list):
                    value = [value]

                if len(value) == size:
                    if dtype == "string":
                        g.edge_properties[name].set_2d_array(
                            np.asarray(value, object))
                    else:
                        g.edge_properties[name].a = np.asarray(value, dtype)
                else:
                    raise ValueError("A list or a np.array with one entry per "
                                 "edge in the graph is required")
        else:
            raise InvalidArgument("Attribute does not exist yet, use "
                                  "set_attribute to create it.")

        self._num_values_set[name] = len(value)

    def set_attribute(self, name, values, edges=None, last_edges=False):
        '''
        Set the edge property.

        Parameters
        ----------
        name : str
            Name of the edge property.
        values : array
            Values that should be set.
        edges : array-like, optional (default: None)
            Edges for which the value of the property should be set. If `edges`
            is not None, it must be an array of shape `(len(values), 2)`.
        '''
        g = self.parent()._graph

        num_edges = g.num_edges()
        num_e     = len(edges) if edges is not None else num_edges

        if num_e != len(values):
            raise ValueError(
                "`edges` and `values` must have the same  size; got "
                "respectively " + str(num_e) + " and " + str(len(values))
                + " entries.")

        if edges is None:
            self[name] = values
        else:
            if last_edges and not self._edges_deleted:
                g.edge_properties[name].a[-num_e:] = values
            else:
                Edge = g.edge
                for e, val in zip(edges, values):
                    gt_e = Edge(*e)
                    g.edge_properties[name][gt_e] = val

        if num_e:
            self._num_values_set[name] = num_edges

    def new_attribute(self, name, value_type, values=None, val=None):
        g = self.parent()._graph

        num_edges = g.num_edges()

        if values is None and val is None:
            self._num_values_set[name] = num_edges

        if val is None:
            if value_type == "int":
                val = int(0)
            elif value_type == "double":
                val = np.NaN
            elif value_type == "string":
                val = ""
            else:
                val = None

        if values is None:
            values = _to_np_array(
                [deepcopy(val) for _ in range(num_edges)],
                value_type)

        if len(values) != num_edges:
            self._num_values_set[name] = 0
            raise ValueError("A list or a np.array with one entry per "
                             "edge in the graph is required")

        # store name and value type in the dict
        super(_GtEProperty, self).__setitem__(name, value_type)

        # store the real values in the attribute
        eprop = g.new_edge_property(value_type, vals=values)
        g.edge_properties[name] = eprop
        self._num_values_set[name] = len(values)

    def edges_deleted(self):
        ''' Notify that some edges were deleted '''
        g = self.parent()

        for key in self:
            self._num_values_set[key] = g.edge_nb()

        self._edges_deleted = True


# ----- #
# Graph #
# ----- #

class _GtGraph(GraphInterface):

    '''
    Subclass of :class:`gt.Graph` that (with
    :class:`~nngt.core._SnapGraph`) unifies the methods to work with either
    `graph-tool` or `SNAP`.
    '''

    _nattr_class = _GtNProperty
    _eattr_class = _GtEProperty

    #-------------------------------------------------------------------------#
    # Constructor and instance properties

    def __init__(self, nodes=0, copy_graph=None, weighted=True, directed=True,
                 **kwargs):
        '''
        @todo: document that
        see :class:`gt.Graph`'s constructor '''
        self._nattr = _GtNProperty(self)
        self._eattr = _GtEProperty(self)

        self._edges_deleted = False

        g = copy_graph.graph if copy_graph is not None else None

        if g is not None:
            from graph_tool import Graph as GtGraph
            from graph_tool.stats import remove_parallel_edges

            num_edges = copy_graph.edge_nb()

            if copy_graph._edges_deleted:
                # set edge filter for non-deleted edges
                eprop = g.new_edge_property(
                    "bool", vals=np.ones(num_edges, dtype=bool))

                g.set_edge_filter(eprop)
                g = GtGraph(g, directed=g.is_directed(), prune=True)

            if not directed and g.is_directed():
                g = g.copy()
                g.set_directed(False)
                remove_parallel_edges(g)
            elif directed and not g.is_directed():
                g = g.copy()
                g.set_directed(True)

            self._from_library_graph(g, copy=True)

            # make edge id property map
            if "eid" in g.edge_properties:
                g.edge_properties["eid"].a = list(range(num_edges))
            else:
                eids = self._graph.new_edge_property(
                    "int", vals=list(range(self._max_eid)))

                g.edge_properties["eid"] = eids

            self._max_eid = num_edges
        else:
            self._graph = nngt._config["graph"](directed=directed)

            if nodes:
                self._graph.add_vertex(nodes)

            # make edge id property map
            self._max_eid = 0

            eids = self._graph.new_edge_property("int")

            self._graph.edge_properties["eid"] = eids

    #-------------------------------------------------------------------------#
    # Graph manipulation

    def edge_id(self, edge):
        '''
        Return the ID a given edge or a list of edges in the graph.
        Raises an error if the edge is not in the graph or if one of the
        vertices in the edge is nonexistent.

        Parameters
        ----------
        edge : 2-tuple or array of edges
            Edge descriptor (source, target).

        Returns
        -------
        index : int or array of ints
            Index of the given `edge`.
        '''
        g = self._graph

        if nonstring_container(edge) and len(edge):
            if is_integer(edge[0]):
                return g.edge_index[g.edge(*edge)]
            elif nonstring_container(edge[0]):
                Edge = g.edge
                return [g.edge_index[Edge(*e)] for e in edge]

        raise AttributeError("`edge` must be either a 2-tuple of ints or "
                             "an array of 2-tuples of ints.")

    def has_edge(self, edge):
        '''
        Whether `edge` is present in the graph.

        .. versionadded:: 2.0
        '''
        try:
            e = self._graph.edge(*edge)

            if e is None:
                return False

            return True
        except:
            return False

    @property
    def edges_array(self):
        '''
        Edges of the graph, sorted by order of creation, as an array of
        2-tuple.
        '''
        g = self._graph

        edges = g.get_edges([g.edge_properties["eid"]])

        order = np.argsort(edges[:, 2])

        return edges[order, :2]

    def _get_edges(self, source_node=None, target_node=None):
        '''
        Called by Graph.get_edges if source_node and target_node are not both
        integers.
        '''
        g = self._graph

        edges = set()

        if source_node is not None:
            if target_node is None:
                if is_integer(source_node):
                    if g.is_directed():
                        return [
                            tuple(e) for e in g.iter_out_edges(source_node)
                        ]

                    return [
                        tuple(e) if e[0] < e[1] else tuple(e[::-1])
                        for e in g.iter_all_edges(source_node)
                    ]

                for s in source_node:
                    if g.is_directed():
                        edges.update((tuple(e) for e in g.iter_out_edges(s)))
                    else:
                        for e in g.iter_all_edges(s):
                            edges.add(
                                tuple(e) if e[0] <= e[1] else tuple(e[::-1]))
            else:
                target_node = {target_node} if is_integer(target_node) \
                              else set(target_node)

                if is_integer(source_node):
                    if g.is_directed():
                        return [tuple(e) for e in g.get_out_edges(source_node)
                                if e[1] in target_node]

                    return [tuple(e) for e in g.get_all_edges(source_node)
                            if e[0] in target_node or e[1] in target_node]

                for s in source_node:
                    if g.is_directed():
                        edges.update((tuple(e) for e in g.iter_out_edges(s)
                                      if e[1] in target_node))
                    else:
                        for e in g.iter_all_edges(s):
                            e = tuple(e) if e[0] <= e[1] else tuple(e[::-1])

                            if e[0] in target_node or e[1] in target_node:
                                edges.add(e)

            return list(edges)

        if target_node is None:
            # return all edges
            return list(g.get_edges())

        if is_integer(target_node):
            if g.is_directed():
                return [tuple(e) for e in g.iter_in_edges(target_node)]

            return [
                tuple(e) if e[0] <= e[1] else tuple(e[::-1])
                for e in g.iter_all_edges(target_node)
            ]

        for t in target_node:
            if g.is_directed():
                edges.update((tuple(e) for e in g.iter_in_edges(t)))
            else:
                for e in g.iter_all_edges(t):
                    edges.add(tuple(e) if e[0] <= e[1] else tuple(e[::-1]))

        return list(edges)

    def new_node(self, n=1, neuron_type=1, attributes=None, value_types=None,
                 positions=None, groups=None):
        '''
        Adding a node to the graph, with optional properties.

        Parameters
        ----------
        n : int, optional (default: 1)
            Number of nodes to add.
        neuron_type : int, optional (default: 1)
            Type of neuron (1 for excitatory, -1 for inhibitory)
        attributes : dict, optional (default: None)
            Dictionary containing the attributes of the nodes.
        value_types : dict, optional (default: None)
            Dict of the `attributes` types, necessary only if the `attributes`
            do not exist yet.
        positions : array of shape (n, 2), optional (default: None)
            Positions of the neurons. Valid only for
            :class:`~nngt.SpatialGraph` or :class:`~nngt.SpatialNetwork`.
        groups : str, int, or list, optional (default: None)
            :class:`~nngt.core.NeuralGroup` to which the neurons belong. Valid
            only for :class:`~nngt.Network` or :class:`~nngt.SpatialNetwork`.

        Returns
        -------
        The node or a list of the nodes created.
        '''
        nodes = self._graph.add_vertex(n)
        nodes = [int(nodes)] if n == 1 else [int(node) for node in nodes]

        attributes = {} if attributes is None else deepcopy(attributes)

        if attributes:
            for k, v in attributes.items():
                v = deepcopy(v)
                if k not in self._nattr:
                    self._nattr.new_attribute(k, value_types[k], val=v)
                else:
                    v = v if nonstring_container(v) else [v]
                    self._nattr.set_attribute(k, v, nodes=nodes)

        # set default values for double attributes that were not set
        # (others are properly handled automatically)
        for k in self.node_attributes:
            if k not in attributes and self.get_attribute_type(k) == "double":
                self.set_node_attribute(k, nodes=nodes, val=np.NaN)

        if self.is_spatial():
            old_pos      = self._pos
            self._pos    = np.full((self.node_nb(), 2), np.NaN)
            num_existing = len(old_pos) if old_pos is not None else 0
            if num_existing != 0:
                self._pos[:num_existing, :] = old_pos

        if positions is not None:
            assert self.is_spatial(), \
                "`positions` argument requires a SpatialGraph/SpatialNetwork."
            self._pos[nodes] = positions

        if groups is not None:
            assert self.is_network(), \
                "`positions` argument requires a Network/SpatialNetwork."
            if nonstring_container(groups):
                assert len(groups) == n, "One group per neuron required."
                for g, node in zip(groups, nodes):
                    self.population.add_to_group(g, node)
            else:
                self.population.add_to_group(groups, nodes)

        if n == 1:
            return nodes[0]
        return nodes

    def delete_nodes(self, nodes):
        '''
        Remove nodes (and associated edges) from the graph.
        '''
        old_enum = self.edge_nb()

        if nonstring_container(nodes):
            for v in reversed(sorted(nodes)):
                self._graph.remove_vertex(v)
        else:
            self._graph.remove_node(nodes)

        for key in self._nattr:
            self._nattr._num_values_set[key] = self.node_nb()

        # tell eattr
        if old_enum != self.edge_nb():
            self._eattr.edges_deleted()
            self._edges_deleted = True

        # check spatial and structure properties
        _post_del_update(self, nodes)

    def new_edge(self, source, target, attributes=None, ignore=False,
                 self_loop=False):
        '''
        Adding a connection to the graph, with optional properties.

        .. versionchanged :: 2.0
            Added `self_loop` argument to enable adding self-loops.

        Parameters
        ----------
        source : :class:`int/node`
            Source node.
        target : :class:`int/node`
            Target node.
        attributes : :class:`dict`, optional (default: ``{}``)
            Dictionary containing optional edge properties. If the graph is
            weighted, defaults to ``{"weight": 1.}``, the unit weight for the
            connection (synaptic strength in NEST).
        ignore : bool, optional (default: False)
            If set to True, ignore attempts to add an existing edge and accept
            self-loops; otherwise an error is raised.
        self_loop : bool, optional (default: False)
            Whether to allow self-loops or not.

        Returns
        -------
        The new connection or None if nothing was added.
        '''
        g = self._graph

        attributes = {} if attributes is None else deepcopy(attributes)

        _set_default_edge_attributes(self, attributes, num_edges=1)

        # check that the edge does not already exist and that nodes are valid
        try:
            edge = g.edge(source, target)
        except ValueError:
            raise InvalidArgument("`source` or `target` does not exist.")

        if edge is None:
            if source == target:
                if not ignore and not self_loop:
                    raise InvalidArgument("Trying to add a self-loop.")
                elif ignore:
                    _log_message(logger, "INFO",
                                 "Self-loop on {} ignored.".format(source))

                    return None

            g.add_edge(source, target, add_missing=False)

            # check distance
            _set_dist_new_edges(attributes, self, [(source, target)])

            # set the attributes
            self._attr_new_edges([(source, target)], attributes=attributes)

            # set edge id
            g.edge_properties["eid"][g.edge(source, target)] = self._max_eid

            self._max_eid += 1
        else:
            if not ignore:
                raise InvalidArgument("Trying to add existing edge.")

            _log_message(logger, "INFO",
                         "Existing edge {} ignored.".format((source, target)))

            return None

        return (source, target)

    def new_edges(self, edge_list, attributes=None, check_duplicates=False,
                  check_self_loops=True, check_existing=True,
                  ignore_invalid=False):
        '''
        Add a list of edges to the graph.

        .. versionchanged:: 2.0
            Can perform all possible checks before adding new edges via the
            ``check_duplicates`` ``check_self_loops``, and ``check_existing``
            arguments.

        Parameters
        ----------
        edge_list : list of 2-tuples or np.array of shape (edge_nb, 2)
            List of the edges that should be added as tuples (source, target)
        attributes : :class:`dict`, optional (default: ``{}``)
            Dictionary containing optional edge properties. If the graph is
            weighted, defaults to ``{"weight": ones}``, where ``ones`` is an
            array the same length as the `edge_list` containing a unit weight
            for each connection (synaptic strength in NEST).
        check_duplicates : bool, optional (default: False)
            Check for duplicate edges within `edge_list`.
        check_self_loops : bool, optional (default: True)
            Check for self-loops.
        check_existing : bool, optional (default: True)
            Check whether some of the edges in `edge_list` already exist in the
            graph or exist multiple times in `edge_list` (also performs
            `check_duplicates`).
        ignore_invalid : bool, optional (default: False)
            Ignore invalid edges: they are not added to the graph and are
            silently dropped. Unless this is set to true, an error is raised
            whenever one of the three checks fails.

        .. warning::

            Setting `check_existing` to False will lead to undefined behavior
            if existing edges are provided! Only use it (for speedup) if you
            are sure that you are indeed only adding new edges.

        Returns
        -------
        Returns new edges only.
        '''
        g = self._graph

        attributes = {} if attributes is None else deepcopy(attributes)
        num_edges  = len(edge_list)
        num_nodes  = self.node_nb()

        # check that all nodes exist
        if num_edges:
            if np.max(edge_list) >= num_nodes:
                raise InvalidArgument("Some nodes do no exist.")

            # set default values for attributes that were not passed
            _set_default_edge_attributes(self, attributes, num_edges)

            # check edges
            new_attr = None

            if check_duplicates or check_self_loops or check_existing:
                edge_list, new_attr = _cleanup_edges(
                    self, edge_list, attributes, check_duplicates,
                    check_self_loops, check_existing, ignore_invalid)
            else:
                edge_list = np.asarray(edge_list)
                new_attr  = attributes

            # check distance
            _set_dist_new_edges(new_attr, self, edge_list)

            # create the edges
            if len(edge_list):
                g.add_edge_list(edge_list)

                # call parent function to set the attributes
                self._attr_new_edges(edge_list, attributes=new_attr)

                # set edge id
                n_e = len(edge_list)

                if self._edges_deleted:
                    Edge = g.edge

                    for e in edge_list:
                        g.edge_properties["eid"][Edge(*e)] = self._max_eid

                        self._max_eid += 1
                else:
                    g.edge_properties["eid"].a[-n_e:] = \
                        list(range(self._max_eid, self._max_eid + n_e))

                    self._max_eid += n_e

        return edge_list

    def delete_edges(self, edges):
        ''' Remove a list of edges '''
        g = self._graph

        Edge = g.edge

        if nonstring_container(edges[0]):
            # fast loop
            [self._graph.remove_edge(Edge(*e)) for e in edges]
        else:
            self._graph.remove_edge(Edge(*edges))

        if len(edges):
            self._eattr.edges_deleted()

            self._edges_deleted = True

    def clear_all_edges(self):
        ''' Remove all edges from the graph '''
        self._graph.clear_edges()
        self._eattr.clear()

    #-------------------------------------------------------------------------#
    # Getters

    def node_nb(self):
        ''' Number of nodes in the graph '''
        return self._graph.num_vertices()

    def edge_nb(self):
        ''' Number of edges in the graph '''
        return self._graph.num_edges()

    def get_degrees(self, mode="total", nodes=None, weights=None):
        w = _get_gt_weights(self, weights)

        if not self._graph.is_directed():
            mode = "total"

        if nodes is not None:
            return self._graph.degree_property_map(
                mode, weight=w).a[nodes]

        return self._graph.degree_property_map(mode, weight=w).a.flatten()

    def is_connected(self, mode="strong"):
        '''
        Return whether the graph is connected.

        Parameters
        ----------
        mode : str, optional (default: "strong")
            Whether to test connectedness with directed ("strong") or
            undirected ("weak") connections.
        '''
        from graph_tool.topology import label_components

        directed  = True if mode == "strong" else False
        directed *= self._graph.is_directed()

        _, hist = label_components(self._graph, directed=directed)

        return len(hist) == 1

    def neighbours(self, node, mode="all"):
        '''
        Return the neighbours of `node`.

        Parameters
        ----------
        node : int
            Index of the node of interest.
        mode : string, optional (default: "all")
            Type of neighbours that will be returned: "all" returns all the
            neighbours regardless of directionality, "in" returns the
            in-neighbours (also called predecessors) and "out" retruns the
            out-neighbours (or successors).

        Returns
        -------
        neighbours : set
            The neighbours of `node`.
        '''
        g = self._graph
        v = g.vertex(node)

        if mode == "all" or not g.is_directed():
            return set(int(n) for n in v.all_neighbours())
        elif mode == "in":
            return set(int(n) for n in v.in_neighbours())
        elif mode == "out":
            return set(int(n) for n in v.out_neighbours())
        else:
            raise ArgumentError('''Invalid `mode` argument {}; possible values
                                are "all", "out" or "in".'''.format(mode))

    def _from_library_graph(self, graph, copy=True):
        ''' Initialize `self._graph` from existing library object. '''
        nodes = graph.num_vertices()
        edges = graph.num_edges()

        if copy:
            self._graph = nngt._config["graph"](g=graph)
        else:
            self._graph = graph

        # get attributes names and "types" and initialize them
        if nodes:
            for key, val in graph.vertex_properties.items():
                try:
                    super(type(self._nattr), self._nattr).__setitem__(
                        key, _get_dtype(val.a[0]))
                except TypeError:
                    super(type(self._nattr), self._nattr).__setitem__(
                        key, _get_dtype(val.get_2d_array([0])))

        if edges:
            for key, val in graph.edge_properties.items():
                if key != 'eid':
                    if val.value_type() == 'string':
                        super(type(self._eattr), self._eattr).__setitem__(
                            key, 'string')
                    elif val.value_type().startswith('vector'):
                        super(type(self._eattr), self._eattr).__setitem__(
                            key, 'object')
                    else:
                        super(type(self._eattr), self._eattr).__setitem__(
                            key, _get_dtype(val.a[0]))

        # make edge ids
        eids = self._graph.new_edge_property(
            "int", vals=list(range(self.edge_nb())))

        self._graph.edge_properties["eid"] = eids



#-*- coding:utf-8 -*-
#
# io/__init__.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
The Input/output (I/O) module contains function to read and write network
files.
"""



#-*- coding:utf-8 -*-
#
# This file is part of the NNGT project to generate and analyze
# neuronal networks and their activity.
# Copyright (C) 2015-2020  Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Plot the degree distributions of a graph
========================================
"""

import nngt
import nngt.plot as nplt

import matplotlib.pyplot as plt

plt.rcParams.update({
    'axes.edgecolor': 'grey', 'xtick.color': 'grey', 'ytick.color': 'grey',
    "figure.facecolor": (0, 0, 0, 0), "axes.facecolor": (0, 0, 0, 0),
    "axes.labelcolor": "grey", "axes.titlecolor": "grey", "text.color": "grey",
    "legend.facecolor": "none"
})


nngt.seed(0)


# %%
# First, let's create a scale-free network

g = nngt.generation.random_scale_free(2.1, 3.2, nodes=1000, avg_deg=100)


# %%
# Plot the degree distribution

nplt.degree_distribution(g, deg_type=["in", "out"], show=True)

# %%
# It's not bad... but we don't see much! Let's move a more relevant scale

nplt.degree_distribution(g, deg_type=["in", "out"], logy=True, show=True)


# %%
# Or we can use Bayesian binning

nplt.degree_distribution(g, deg_type=["in", "out"], num_bins="bayes",
                         show=True)



#-*- coding:utf-8 -*-
#
# plot/plt_properties.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Tools to plot graph properties """

import numpy as np

import nngt
from nngt.lib import InvalidArgument, nonstring_container
from nngt.analysis import (degree_distrib, betweenness_distrib,
                           node_attributes, binning)
from .custom_plt import palette_continuous, palette_discrete, format_exponent


__all__ = [
    'degree_distribution',
    'betweenness_distribution',
    'edge_attributes_distribution',
    'node_attributes_distribution',
    'compare_population_attributes',
    "correlation_to_attribute",
]


# ---------------------- #
# Plotting distributions #
# ---------------------- #

def degree_distribution(network, deg_type="total", nodes=None,
                        num_bins='doane', weights=False, logx=False,
                        logy=False, axis=None, axis_num=None, colors=None,
                        norm=False, show=True, title=None, **kwargs):
    '''
    Plotting the degree distribution of a graph.

    Parameters
    ----------
    graph : :class:`~nngt.Graph` or subclass
        The graph to analyze.
    deg_type : string or N-tuple, optional (default: "total")
        Type of degree to consider ("in", "out", or "total")
    nodes : list or numpy.array of ints, optional (default: all nodes)
        Restrict the distribution to a set of nodes.
    num_bins : str, int or N-tuple, optional (default: 'doane'):
        Number of bins used to sample the distribution. Defaults to 'doane'.
        Use to 'auto' for numpy automatic selection or 'bayes' for unsupervised
        Bayesian blocks method.
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then use binary edges; if ``True``, uses the 'weight' edge attribute,
        otherwise uses any valid edge attribute required.
    logx : bool, optional (default: False)
        Use log-spaced bins.
    logy : bool, optional (default: False)
        Use logscale for the degree count.
    axis : :class:`matplotlib.axes.Axes` instance, optional (default: new one)
        Axis which should be used to plot the histogram, if None, a new one is
        created.
    show : bool, optional (default: True)
        Show the Figure right away if True, else keep it warm for later use.
    **kwargs : keyword arguments for :func:`matplotlib.axes.Axes.bar`.
    '''
    import matplotlib.pyplot as plt

    if axis is None:
        fig, axis = plt.subplots()

    empty_axis = axis.has_data()

    axis.axis('tight')

    if "alpha" not in kwargs:
        kwargs["alpha"] = 1 if isinstance(deg_type, str) else 0.5

    labels = kwargs.get('label', None)

    if not nonstring_container(labels) and labels is not None:
        labels = [labels]

    # get degrees
    mincounts, maxcounts, allbins = np.inf, 0, []

    if isinstance(deg_type, str):
        counts, bins = degree_distrib(network, deg_type, nodes,
                                      weights, logx, num_bins)
        bottom_count = -1 if logy else 0
        if norm:
            counts = counts / float(np.sum(counts))
        if logy:
            counts = counts.astype(float)
            counts[counts<1] = 0.1
            counts = np.log(counts)

        max_nnz = np.where(counts > 0)[0][-1]
        maxcounts, mincounts = counts.max(), np.min(counts[counts > 0])
        allbins.extend(bins)

        if "label" not in kwargs:
            kwargs["label"] = deg_type[0].upper() + deg_type[1:] + " degree"
        axis.bar(
            bins[:-1], counts, np.diff(bins), bottom=bottom_count, **kwargs)
    else:
        if colors is None:
            colors = palette_continuous(np.linspace(0.,0.5, len(deg_type)))

        if not nonstring_container(num_bins):
            num_bins = [num_bins for _ in range(len(deg_type))]

        for i, s_type in enumerate(deg_type):
            counts, bins = degree_distrib(network, s_type, nodes,
                                          weights, logx, num_bins[i])

            bottom_count = -1 if logy else 0

            if norm:
                counts = counts / float(np.sum(counts))
            if logy:
                counts = counts.astype(float)
                counts[counts<1] = 0.1
                counts = np.log(counts) + 1

            maxcounts_tmp = counts.max()
            mincounts_tmp = np.min(counts[counts>0])
            
            mincounts = min(mincounts, mincounts_tmp)
            maxcounts = max(maxcounts, maxcounts_tmp)

            allbins.extend(bins)

            if labels is None:
                kwargs['label'] = s_type[0].upper() + s_type[1:] + " degree"
            else:
                kwargs['label'] = labels[i]

            axis.bar(
                bins[:-1], counts, np.diff(bins), color=colors[i],
                bottom=bottom_count, align='edge', **kwargs)

    axis.set_xlabel("Degree")
    axis.set_ylabel("Node count")
    title_start = (deg_type[0].upper() + deg_type[1:] + '-d'
                   if isinstance(deg_type, str) else 'D')
    if title != "":
        str_title = title
        if title is None:
            str_title = "{}egree distribution for {}".format(
                title_start, network.name)
        axis.set_title(str_title, x=0., y=1.05, loc='left')

    # restore ylims and xlims and adapt if necessary
    _set_scale(axis, np.array(allbins), mincounts, maxcounts, logx, logy)

    plt.legend()

    if show:
        plt.show()


def attribute_distribution(network, attribute, num_bins='auto', logx=False,
                           logy=False, axis=None, norm=False, show=True):
    '''
    Plotting the distribution of a graph attribute (e.g. "weight", or
    "distance" is the graph is spatial).

    Parameters
    ----------
    graph : :class:`~nngt.Graph` or subclass
        Graph to analyze.
    attribute : string or tuple of strings
        Name of a graph attribute.
    num_bins : int or 'auto', optional (default: 'auto'):
        Number of bins used to sample the distribution. Defaults to
        unsupervised Bayesian blocks method.
    logx : bool, optional (default: False)
        use log-spaced bins.
    logy : bool, optional (default: False)
        use logscale for the degree count.
    axis : :class:`matplotlib.axis.Axis` instance, optional (default: new one)
        Axis which should be used to plot the histogram, if None, a new one is
        created.
    show : bool, optional (default: True)
        Show the Figure right away if True, else keep it warm for later use.
    '''
    import matplotlib.pyplot as plt
    if axis is None:
        axis = plt.gca()

    mincounts, maxcounts, bins = 0, 0, None

    if isinstance(attribute, str):
        values = network.get_edge_attributes(name=attribute)
        counts, bins = _hist(
            values, num_bins, norm, logx, attribute, axis, **kwargs)
        maxcounts, mincounts = counts.max(), np.min(counts[counts>0])
        maxbins, minbins     = bins.max(), bins.min()
    else:
        raise NotImplementedError("Multiple attribute plotting not ready yet")
        #~ colors = palette(np.linspace(0.,0.5,len(deg_type)))
        #~ m = ["o", "s", "D", "x"]
        #~ lines, legends = [], []
        #~ for i,s_type in enumerate(deg_type):
            #~ counts,bins = degree_distrib(network, s_type, nodes,
                                         #~ weights, logx, num_bins)
            #~ maxcounts_tmp,mincounts_tmp = counts.max(),counts.min()
            #~ maxbins_tmp,minbins_tmp = bins.max(),bins.min()
            #~ maxcounts = max(maxcounts,maxcounts_tmp)
            #~ maxbins = max(maxbins,maxbins_tmp)
            #~ minbins = min(minbins,minbins_tmp)
            #~ lines.append(ax1.scatter(bins, counts, c=colors[i], marker=m[i]))
            #~ legends.append(attribute)
        #~ ax1.legend(lines, legends)
    if nngt._config['use_tex']:
        axis.set_xlabel(attribute.replace("_", "\\_"))

    axis.set_ylabel("Node count")

    _set_scale(ax1, bins, mincounts, maxcounts, logx, logy)

    axis.set_title(
        "Attribute distribution for {}".format(network.name), x=0., y=1.05,
        loc='left')

    plt.legend()

    axis.axis('tight')

    if show:
        plt.show()


def betweenness_distribution(
    network, btype="both", weights=False, nodes=None, logx=False, logy=False,
    num_nbins=None, num_ebins=None, axes=None, colors=None, norm=False,
    legend_location='right', show=True, **kwargs):
    '''
    Plotting the betweenness distribution of a graph.

    Parameters
    ----------
    graph : :class:`~nngt.Graph` or subclass
        the graph to analyze.
    btype : string, optional (default: "both")
        type of betweenness to display ("node", "edge" or "both")
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then use binary edges; if ``True``, uses the 'weight' edge attribute,
        otherwise uses any valid edge attribute required.
    nodes : list or numpy.array of ints, optional (default: all nodes)
        Restrict the distribution to a set of nodes (taken into account only
        for the node attribute).
    logx : bool, optional (default: False)
        use log-spaced bins.
    logy : bool, optional (default: False)
        use logscale for the degree count.
    num_nbins : int or 'auto', optional (default: None):
        Number of bins used to sample the node distribution. Defaults to
        `max(num_nodes / 50., 10)`.
    num_ebins : int or 'auto', optional (default: None):
        Number of bins used to sample the edge distribution. Defaults to
        `max(num_edges / 500., 10)` ('auto' method will be slow).
    axes : list of :class:`matplotlib.axis.Axis`, optional (default: new ones)
        Axes which should be used to plot the histogram, if None, new ones are
        created.
    legend_location : str, optional (default; 'right')
        Location of the legend.
    show : bool, optional (default: True)
        Show the Figure right away if True, else keep it warm for later use.
    '''
    import matplotlib.pyplot as plt

    if btype not in ("node", "edge", "both"):
        raise InvalidArgument('`btype` must be one of the following: '
                              '"node", "edge", "both".')

    num_axes = 2 if btype == "both" else 1

    # create new axes or get them from existing ones
    ax = None

    if axes is None:
        fig, ax = plt.subplots()
        fig.patch.set_visible(False)
        ax.grid(False, axis='y')
        axes = [ax.twinx()]
        if num_axes == 2:
            ax2 = ax.twiny()
            axes.append(ax2)
            ax.grid(False, axis='x')
            ax.yaxis.tick_right()
            ax.yaxis.set_label_position("right")
        else:
            ax.set_yticks([])
    else:
        ax = axes[0]

    ax1 = axes[0]

    ax1.yaxis.tick_left()
    ax1.yaxis.set_label_position("left")
    ax2 = axes[-1]

    # get betweenness
    if num_ebins is None:
        num_ebins = int(max(network.edge_nb() / 500., 10))

    if num_nbins is None:
        num_nbins = int(max(network.node_nb() / 50., 10))

    ncounts, nbins, ecounts, ebins = betweenness_distrib(
        network, weights, nodes=nodes, num_nbins=num_nbins,
        num_ebins=num_ebins, log=logx)

    if norm:
        ncounts = ncounts / float(np.sum(ncounts))
        ecounts = ecounts / np.sum(ecounts)

    if colors is None:
        colors = palette_continuous(np.linspace(0.05, 0.5, 2))

    alpha = kwargs.get("alpha", None)

    if alpha is None:
        kwargs["alpha"] = 0.5 if btype == "both" else 1

    
    x = 0 if legend_location == 'left' else 1

    # plot
    if btype in ("node", "both"):
        ax1.bar(
            nbins[:-1], ncounts, np.diff(nbins), color=colors[0], align='edge',
            **kwargs)

        ax1.legend(
            ["Node betweenness"], bbox_to_anchor=[x, 1],
            loc='upper ' + legend_location)

        ax.set_xlabel("Node betweenness")
        ax1.set_ylabel("Node count")
        ax1.ticklabel_format(axis='x', style='sci', scilimits=(-3, 2))

        _set_scale(ax1, nbins, np.min(ncounts[ncounts>0]), ncounts.max(),
                   logx, logy)
    if btype in ("edge", "both"):
        ax2.bar(
            ebins[:-1], ecounts, np.diff(ebins), color=colors[-1],
            align='edge', **kwargs)

        ax2.legend(
            ["Edge betweenness"], bbox_to_anchor=[x, 1],
            loc='upper ' + legend_location)

        ax2.set_xlim([ebins.min(), ebins.max()])
        ax.set_ylim([0, 1.1*ecounts.max()])
        ax2.set_xlabel("Edge betweenness")
        ax.set_ylabel("Edge count")
        ax2.ticklabel_format(axis='x', style='sci', scilimits=(-3, 2))

        _set_scale(ax2, ebins, np.min(ecounts[ecounts>0]), ecounts.max(),
                   logx, logy)
    if btype == "both":
        ax2.legend(
            ["Edge betweenness"], bbox_to_anchor=[x, 0.88],
            loc='upper ' + legend_location)
        ax1.legend(
            ["Node betweenness"], bbox_to_anchor=[x, 0.88],
            loc='lower ' + legend_location)
        #~ ax1.spines['top'].set_color('none')
        #~ ax1.spines['right'].set_color('none')
        plt.subplots_adjust(top=0.85)
        #~ ax.xaxis.set_label_position("top")
        ax1.grid(False)
        ax2.grid(False)
        if not logx:
            ax2 = format_exponent(ax2, 'x', (1., 1.1))
            ax1 = format_exponent(ax1, 'x', (1., -0.05))

    y = 1.15 if btype == "both" else 1.05 

    ax1.set_title(
        "Betweenness distribution for {}".format(network.name), x=0., y=y,
        loc='left')

    # ~ ax1.axis('tight')
    # ~ ax2.axis('tight')
    plt.tight_layout()

    if show:
        plt.show()


# ------------------------ #
# Plotting node attributes #
# ------------------------ #

def node_attributes_distribution(network, attributes, nodes=None,
                                 num_bins='auto', logx=False, logy=False,
                                 norm=False, title=None, colors=None,
                                 show=True, **kwargs):
    '''
    Return node `attributes` for a set of `nodes`.

    Parameters
    ----------
    network : :class:`~nngt.Graph`
        The graph where the `nodes` belong.
    attributes : str or list
        Attributes which should be returned, among:
        * "betweenness"
        * "clustering"
        * "closeness"
        * "in-degree", "out-degree", "total-degree"
        * "subgraph_centrality"
        * "b2" (requires NEST)
        * "firing_rate" (requires NEST)
    nodes : list, optional (default: all nodes)
        Nodes for which the attributes should be returned.
    num_bins : int or list, optional (default: 'auto')
        Number of bins to plot the distributions. If only one int is provided,
        it is used for all attributes, otherwise a list containing one int per
        attribute in `attributes` is required. Defaults to unsupervised
        Bayesian blocks method.
    logx : bool or list, optional (default: False)
        Use log-spaced bins.
    logy : bool or list, optional (default: False)
        use logscale for the node count.
    '''
    import matplotlib.pyplot as plt

    if not nonstring_container(attributes):
        attributes = [attributes]
    else:
        attributes = [name for name in attributes]

    num_attr = len(attributes)
    num_bins = _format_arg(num_bins, num_attr, 'num_bins')
    colors = _format_arg(colors, num_attr, 'num_bins')
    logx = _format_arg(logx, num_attr, 'logx')
    logy = _format_arg(logy, num_attr, 'logy')
    num_plot = 0

    # kwargs that will not be passed:
    ignore = ["degree", "betweenness"] + attributes
    new_kwargs = {k: v for k, v in kwargs.items() if k not in ignore}
    fig = None
    if new_kwargs == kwargs:
        fig = plt.figure()
    else:
        fig = plt.figure(plt.get_fignums()[-1])
    fig.patch.set_visible(False)

    # plot degrees if required
    degrees = []
    for name in attributes:
        if "degree" in name.lower():
            degrees.append(name[:name.find("-")])

    if degrees:
        # get the indices where a degree-related attribute is required
        indices, colors_deg, logx_deg, logy_deg = [], [], 0, 0
        for i, name in enumerate(attributes):
            if "degree" in name:
                indices.append(i)
                if colors is not None:
                    colors_deg.append(colors[i])
                logx_deg += logx[i]
                logy_deg += logy[i]
        colors_deg = None if colors is None else colors_deg
        indices.sort()
        deg_bin = [num_bins[i] for i in indices]
        for idx in indices[::-1]:
            del num_bins[idx]
            del attributes[idx]
            del logx[idx]
            del logy[idx]
            if colors is not None:
                del colors[idx]
        if "degree" in kwargs:
            degree_distribution(
                network, deg_type=degrees, nodes=nodes, num_bins=deg_bin,
                logx=logx_deg, logy=logy_deg, norm=norm, axis=kwargs["degree"],
                colors=colors_deg, show=False, **new_kwargs)
        else:
            fig, ax = _set_new_plot(
                fignum=fig.number, num_new_plots=1,
                names=['Degree distribution'])
            degree_distribution(
                network, deg_type=degrees, nodes=nodes, num_bins=deg_bin,
                logx=logx_deg, logy=logy_deg, axis=ax[0], colors=colors_deg,
                norm=norm, show=False)
        num_plot += 1

    # plot betweenness if needed
    if "betweenness" in attributes:
        idx = attributes.index("betweenness")
        if "betweenness" in kwargs:
            betweenness_distribution(
                network, btype="node", nodes=nodes, logx=logx[idx],
                logy=logy[idx], axes=kwargs["betweenness"],
                colors=[colors[idx]], norm=norm, show=False, **new_kwargs)
        else:
            fig, axes = _set_new_plot(
                fignum=fig.number, num_new_plots=1,
                names=['Betweenness distribution'])
            betweenness_distribution(
                network, btype="node", nodes=nodes, logx=logx[idx],
                logy=logy[idx], norm=norm, axes=axes, show=False)
        del attributes[idx]
        del num_bins[idx]
        del logx[idx]
        del logy[idx]
        if colors is not None:
            del colors[idx]
        num_plot += 1

    # plot the remaining attributes
    values = node_attributes(network, attributes, nodes=nodes)
    for i, (attr, val) in enumerate(values.items()):
        if attr in kwargs:
            new_kwargs['color'] = colors[i]
            counts, bins = _hist(
                val, num_bins[i], norm, logx[i], attr, kwargs[attr],
                **new_kwargs)
        else:
            fig, ax = _set_new_plot(fignum=fig.number, names=[attr])
            counts, bins = _hist(
                val, num_bins[i], norm, logx[i], attr, ax[0], **kwargs)
            end_attr = attr[1:]
            if nngt._config["use_tex"]:
                end_attr = end_attr.replace("_", "\\_")
            ax[0].set_title("{}{} distribution for {}".format(
                attr[0].upper(), end_attr, network.name), y=1.05)
            ax[0].set_ylabel("Node count")
            ax[0].set_xlabel(attr[0].upper() + end_attr)
            _set_scale(ax[0], bins, np.min(counts[counts>0]),
                       counts.max(), logx[i], logy[i])
        num_plot += 1

    # adjust space, set title, and show
    _format_and_show(fig, num_plot, values, title, show)


def edge_attributes_distribution(network, attributes, edges=None,
                                 num_bins='auto', logx=False, logy=False,
                                 norm=False, title=None, colors=None,
                                 show=True, **kwargs):
    '''
    Return node `attributes` for a set of `nodes`.

    .. versionadded:: 1.0.3

    Parameters
    ----------
    network : :class:`~nngt.Graph`
        The graph where the `nodes` belong.
    attributes : str or list
        Attributes which should be returned (e.g. "betweenness", "delay",
        "weight").
    edges : list, optional (default: all edges)
        Edges for which the attributes should be returned.
    num_bins : int or list, optional (default: 'auto')
        Number of bins to plot the distributions. If only one int is provided,
        it is used for all attributes, otherwise a list containing one int per
        attribute in `attributes` is required. Defaults to unsupervised
        Bayesian blocks method.
    logx : bool or list, optional (default: False)
        Use log-spaced bins.
    logy : bool or list, optional (default: False)
        use logscale for the node count.
    '''
    import matplotlib.pyplot as plt
    if not nonstring_container(attributes):
        attributes = [attributes]
    else:
        attributes = [name for name in attributes]

    num_attr = len(attributes)
    num_bins = _format_arg(num_bins, num_attr, 'num_bins')
    colors   = _format_arg(colors, num_attr, 'num_bins')
    logx     = _format_arg(logx, num_attr, 'logx')
    logy     = _format_arg(logy, num_attr, 'logy')
    num_plot = 0

    # kwargs that will not be passed:
    ignore = ["weight", "delay", "betweenness"] + attributes
    new_kwargs = {k: v for k, v in kwargs.items() if k not in ignore}

    fig = plt.figure()
    fig.patch.set_visible(False)

    # plot betweenness if needed
    if "betweenness" in attributes:
        idx = attributes.index("betweenness")
        fig, axes = _set_new_plot(
                fignum=fig.number, num_new_plots=1,
                names=['Betweenness distribution'])
        betweenness_distribution(
            network, btype="edges", edges=edges, logx=logx[idx],
            logy=logy[idx], norm=norm, axes=axes, show=False)

        del attributes[idx]
        del num_bins[idx]
        del logx[idx]
        del logy[idx]

        if colors is not None:
            del colors[idx]

        num_plot += 1

    # plot the remaining attributes
    for i, attr in enumerate(attributes):
        val = network.get_edge_attributes(edges=edges, name=attr)

        if attr in kwargs:
            new_kwargs['color'] = colors[i]
            counts, bins = _hist(
                val, num_bins[i], norm, logx[i], attr, kwargs[attr],
                **new_kwargs)
        else:
            fig, ax = _set_new_plot(fignum=fig.number, names=[attr])

            counts, bins = _hist(
                val, num_bins[i], norm, logx[i], attr, ax[0], **kwargs)

            end_attr = attr[1:]

            if nngt._config["use_tex"]:
                end_attr = end_attr.replace("_", "\\_")

            ax[0].set_title("{}{} distribution for {}".format(
                attr[0].upper(), end_attr, network.name), y=1.05)
            ax[0].set_ylabel("Node count")
            ax[0].set_xlabel(attr[0].upper() + end_attr)
            _set_scale(ax[0], bins, np.min(counts[counts>0]), counts.max(),
                       logx[i], logy[i])
        num_plot += 1

    # adjust space, set title, and show
    _format_and_show(fig, num_plot, attributes, title, show)


def correlation_to_attribute(network, reference_attribute, other_attributes,
                             attribute_type="node", nodes=None, edges=None,
                             fig=None, title=None, show=True):
    '''
    For each node plot the value of `reference_attributes` against each of the
    `other_attributes` to check for correlations.

    .. versionchanged :: 2.0
        Added `fig` argument.

    Parameters
    ----------
    network : :class:`~nngt.Graph`
        The graph where the `nodes` belong.
    reference_attribute : str or array-like
        Attribute which should serve as reference, among:

        * "betweenness"
        * "clustering"
        * "in-degree", "out-degree", "total-degree"
        * "in-strength", "out-strength", "total-strength"
        * "subgraph_centrality"
        * "b2" (requires NEST)
        * "firing_rate" (requires NEST)
        * a custom array of values, in which case one entry per node in `nodes`
          is required.
    other_attributes : str or list
        Attributes that will be compared to the reference.
    attribute_type : str, optional (default: 'node')
        Whether we are dealing with 'node' or 'edge' attributes
    nodes : list, optional (default: all nodes)
        Nodes for which the attributes should be returned.
    edges : list, optional (default: all edges)
        Edges for which the attributes should be returned.
    fig : :class:`matplotlib.figure.Figure`, optional (default: new Figure)
        Figure to which the plot should be added.
    title : str, optional (default: automatic).
        Custom title, use "" to remove the automatic title.
    show : bool, optional (default: True)
        Whether the plot should be displayed immediately.
    '''
    import matplotlib.pyplot as plt

    if not nonstring_container(other_attributes):
        other_attributes = [other_attributes]

    fig = plt.figure() if fig is None else fig
    fig.patch.set_visible(False)

    # get reference data
    ref_data = reference_attribute

    if isinstance(reference_attribute, str):
        if attribute_type == "node":
            ref_data = node_attributes(network, reference_attribute,
                                       nodes=nodes)
        else:
            ref_data = network.get_edge_attributes(edges=edges,
                                                   name=reference_attribute)
    else:
        reference_attribute = "user defined attribute"

    # plot the remaining attributes
    assert isinstance(other_attributes, (str, list)), \
        "Only attribute names are allowed for `other_attributes`"

    values = {}

    if attribute_type == "node":
        values = node_attributes(network, other_attributes, nodes=nodes)
    else:
        for name in other_attributes:
            values[name] = network.get_edge_attributes(edges=edges, name=name)

    fig, axes = _set_new_plot(fignum=fig.number, names=other_attributes)

    for i, (attr, val) in enumerate(values.items()):
        end_attr = attr[1:]
        end_ref_attr = reference_attribute[1:]

        if nngt._config["use_tex"]:
            end_attr = end_attr.replace("_", "\\_")
            end_ref_attr = end_ref_attr.replace("_", "\\_")

        # reference nodes
        axes[i].plot(val, ref_data, ls="", marker="o")
        axes[i].set_xlabel(attr[0].upper() + end_attr)
        axes[i].set_ylabel(reference_attribute[0].upper() + end_ref_attr)
        axes[i].set_title(
            "{}{} vs {}".format(
                reference_attribute[0].upper(), end_ref_attr, attr[0] + \
                end_attr),
            loc='left', x=0., y=1.05)

    fig.suptitle(network.name)

    plt.tight_layout()

    # adjust space, set title, and show
    _format_and_show(fig, 0, values, title, show)


def compare_population_attributes(network, attributes, nodes=None,
                                  reference_nodes=None, num_bins='auto',
                                  reference_color="gray", title=None,
                                  logx=False, logy=False, show=True, **kwargs):
    '''
    Compare node `attributes` between two sets of nodes. Since number of nodes
    can vary, normalized distributions are used.

    Parameters
    ----------
    network : :class:`~nngt.Graph`
        The graph where the `nodes` belong.
    attributes : str or list
        Attributes which should be returned, among:
        * "betweenness"
        * "clustering"
        * "in-degree", "out-degree", "total-degree"
        * "subgraph_centrality"
        * "b2" (requires NEST)
        * "firing_rate" (requires NEST)
    nodes : list, optional (default: all nodes)
        Nodes for which the attributes should be returned.
    reference_nodes : list, optional (default: all nodes)
        Reference nodes for which the attributes should be returned in order
        to compare with `nodes`.
    num_bins : int or list, optional (default: 'auto')
        Number of bins to plot the distributions. If only one int is provided,
        it is used for all attributes, otherwize a list containing one int per
        attribute in `attributes` is required. Defaults to unsupervised
        Bayesian blocks method.
    logx : bool or list, optional (default: False)
        Use log-spaced bins.
    logy : bool or list, optional (default: False)
        use logscale for the node count.
    '''
    import matplotlib.pyplot as plt
    if not isinstance(reference_color, str):
        raise InvalidArgument("`reference_color` must be a valid matplotlib "
                              "color string.")
    # plot the non reference nodes
    node_attributes_distribution(network, attributes, nodes=nodes,
                                 num_bins=num_bins, logx=logx, logy=logx,
                                 norm=True, title=title, show=False, **kwargs)
    # get the last figure and put the axes to a dict
    # (order is degree, betweenness, attributes)
    fig = plt.figure(plt.get_fignums()[-1])
    fig.patch.set_visible(False)
    axes = fig.get_axes()
    ref_kwargs = kwargs.copy()
    ref_kwargs.update({'alpha': 0.5})
    for ax in axes:
        if ax.name == 'Degree distribution':
            ref_kwargs['degree'] = ax
        elif ax.name == 'Betweenness distribution':
            ref_kwargs['betweenness'] = [ax]  # expect list
        else:
            ref_kwargs[ax.name] = ax
    node_attributes_distribution(
        network, attributes, nodes=reference_nodes, num_bins=num_bins,
        logx=logx, logy=logx, colors=reference_color, norm=True, title=title,
        show=show, **ref_kwargs)


# --------- #
# Histogram #
# --------- #

def _hist(values, num_bins, norm, logx, label, axis, **kwargs):
    '''
    Compute and draw the histogram.

    Returns
    -------
    counts, bins
    '''
    bins = binning(values, bins=num_bins, log=logx)

    counts, bins = np.histogram(values, bins=bins)

    if norm:
        counts = np.divide(counts, float(np.sum(counts)))

    axis.bar(
        bins[:-1], counts, np.diff(bins), label=label, **kwargs)

    if logx:
        axis.set_xscale("log")

    return counts, bins


# ----------------- #
# Figure management #
# ----------------- #

def _set_new_plot(fignum=None, num_new_plots=1, names=None, sharex=None):
    import matplotlib.pyplot as plt
    # get the figure and compute the new number of rows and cols
    fig = plt.figure(num=fignum)
    num_axes = len(fig.axes) + num_new_plots
    if names is not None:
        num_axes = len(fig.axes) + len(names)
        num_new_plots = len(names)
    num_cols = max(int(np.ceil(np.sqrt(num_axes))), 1)
    ratio = num_axes/float(num_cols)
    num_rows = int(ratio)
    if int(ratio) != int(np.ceil(ratio)):
        num_rows += 1
    # change the geometry
    for i in range(num_axes - num_new_plots):
        fig.axes[i].change_geometry(num_rows, num_cols, i+1)
    lst_new_axes = []
    n_old = num_axes-num_new_plots+1
    for i in range(num_new_plots):
        if fig.axes:
            lst_new_axes.append(
                fig.add_subplot(num_rows, num_cols, n_old+i, sharex=sharex))
        else:
            lst_new_axes.append(fig.add_subplot(num_rows, num_cols, n_old+i))
        if names is not None:
            lst_new_axes[-1].name = names[i]
    return fig, lst_new_axes


def _log_format(y, pos):
    '''
    Needed to move log values by one, so first increment, then decrement
    '''
    # rounding err for 4 so add 0.4 to avoid it
    #~ return '{}'.format(int(np.e*np.e**(y-1) + 0.4)) if y > -1 else 0
    return y


def _set_scale(ax1, xbins, mincounts, maxcounts, logx, logy):
    if logx:
        minposbin = xbins[xbins > 0][0]
        ax1.set_xscale("symlog", linthresh=0.2*minposbin)
        next_power = np.ceil(np.log10(xbins.max()))
        ax1.set_xlim([0.8*xbins.min(), 10**next_power])
    else:
        maxbins, minbins = xbins.max(), xbins.min()
        bin_margin = 0.05*(maxbins - minbins)
        if minbins - bin_margin < ax1.get_xlim()[0]:
            ax1.set_xlim(left=(minbins - bin_margin))
        if maxbins + bin_margin > ax1.get_xlim()[1]:
            ax1.set_xlim(right=(maxbins + bin_margin))

    if logy:
        ax1.set_ylim(0, 2*maxcounts)
        ax1.set_yscale("symlog", linthresh=1)
    else:
        if 1.05*maxcounts > ax1.get_ylim()[1]:
            ax1.set_ylim([0, 1.05*maxcounts])


def _set_ax_lims(ax, maxx, minx, maxy, miny, logx=False, logy=False):
    if ax.has_data():
        xlims = ax.get_xlim()
        ylims = ax.get_ylim()
        if not logx:
            Dx = maxx - minx
            minx = minx - 0.01*Dx
            maxx = maxx + 0.01*Dx
            Dy = maxy - miny
            miny = miny - 0.01*Dy
            maxy = maxy + 0.01*Dy
        else:
            minx /= 1.5
            maxx *= 1.5
        if minx > xlims[0]:
            minx = xlims[0]
        if maxx < xlims[1]:
            maxx = xlims[1]
        if miny > ylims[0]:
            miny = ylims[0]
        if maxy < ylims[1]:
            maxy = ylims[1]
    _set_xlim(ax, maxx, minx, logx)
    _set_ylim(ax, maxy, miny, logy)


def _set_xlim(ax, maxx, minx, log):
    if log:
        ax.set_xscale("log")
        ax.set_xlim([max(minx, 1e-10)/1.5, 1.5*maxx])
    else:
        Dx = maxx - minx
        ax.set_xlim([minx - 0.01*Dx, maxx + 0.01*Dx])


def _set_ylim(ax, maxy, miny, log):
    if log:
        ax.set_yscale("log")
        ax.set_ylim([max(miny, 1e-10)/1.5, 1.5*maxy])
    else:
        Dy = maxy - miny
        ax.set_ylim([miny - 0.01*Dy, maxy + 0.01*Dy])


def _format_and_show(fig, num_plot, values, title, show):
    import matplotlib.pyplot as plt
    num_cols = max(int(np.ceil(np.sqrt(num_plot + len(values)))), 1)
    ratio = (num_plot + len(values)) / float(num_cols)
    num_rows = int(ratio)
    if int(ratio) != int(np.ceil(ratio)):
        num_rows += 1
    #~ plt.subplots_adjust(hspace=num_rows*0.2, wspace=num_cols*0.1, left=0.075,
                        #~ right=0.95, top=0.9, bottom=0.075)
    if title is not None:
        fig.suptitle(title)
    if show:
        plt.show()


def _format_arg(arg, num_expected, arg_name):
    if nonstring_container(arg):
        assert len(arg) == num_expected, "One entry per attribute " +\
            "required for `" + arg_name + "`."
    elif arg is not None:
        arg = [arg for _ in range(num_expected)]
    return arg



#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# This file is part of the PyNCulture project, which aims at providing tools to
# easily generate complex neuronal cultures.
# Copyright (C) 2017 SENeC Initiative
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Some tools present in this files are modified from the imagedraw project by
# Philippe Guglielmetti.
#
# Contributors: S. Bottani, A. Quaresima

'''
Read dxf graphic file of growth region boundaries

:requires: `dxfgrabber <http://pypi.python.org/pypi/dxfgrabber/>`

Input
=====
link path to dxf 2D graphic file to read

Output
======
Shapes in as shapely objects.

Lists DX and DY of  x and y homcoord.nates of vectors to draw the graph
i.e. for a square defined by extermities:
top_left:  (x_tl,y_tl)
top_right: (x_tr,y_tr)
bottom_left: (x_bl,y_bl)
bottom_right: (x_br,y_br)
DX= [[x_tl,x_tlr],[x_tr,x_br],[x_br,x_bl],[x_bl,x_tl]]
DY= [[y_tl,y_tlr],[y_tr,y_br],[y_br,y_bl],[y_bl,y_tl]]
'''

import itertools
import logging

import numpy as np

import dxfgrabber

from shapely.geometry import Polygon

from . import homcoord


# ---------------- #
# DFX parser class #
# ---------------- #

class DXF:

    '''
    Class responsible for parsing DXF files
    '''

    def __init__(self, filename, interpolate_curve=50, layers=None,
                 ignore=None):
        """
        Reads a .dxf file
        
        Parameters
        ----------
        filename : str
            Path to .dxf file.
        interpolate_curve : int, optional (default: 50)
            Number of points to use to interpolate a curve.
        layers : list or dict, optional (default: all layers)
            Layers to handle.
        ignore : list, optional (default: None)
             Names of entity types to ignore.
        """
        if ignore is None:
            ignore = []
        self.dxf = dxfgrabber.readfile(filename)
        self.interp = interpolate_curve
        self.layers = layers
        self.ignore = ignore

    def entities(self, ent=None):
        """iterator over dxf or block entities"""
        if not ent:
            ent = self.dxf.entities
        for e in ent:
            if self.layers and (e.layer not in self.layers):
                continue
            elif e.dxftype in self.ignore:
                continue
            else:
                yield e

    def points(self, interpolate_curve=None, ent=None):
        '''
        Lists of segments defining the drawing of the boundaries:

        - for LINES entities these are the extremal points of segments
        - for CIRCLE entities these are extramal points of the sides of a
           polygon approximating the circle with `interpolate_curve` vertices.
        '''
        if interpolate_curve is None:
            interpolate_curve = self.interp
        segments_list = []
        for e in self.entities():
            if e.dxftype == 'LINE':
                segments_list.append(homcoord.segment(
                    homcoord.Pt(e.start[:2]), homcoord.Pt(e.end[:2])))
                logging.warning('entity %s in dxf file' % e)
            elif e.dxftype == 'CIRCLE':
                c = homcoord.Pt(e.center[:2])
                rayon = e.radius
                the_angles = np.linspace(-np.pi, np.pi, interpolate_curve)
                logging.warning('entity %s in dxf file' % e)
                for angle, following_angle in itertools.izip(the_angles,
                                                             the_angles[1:]):
                    Pt1 = c + homcoord.Pt(rayon * np.cos(angle),
                                          rayon * np.sin(angle))
                    Pt2 = c + \
                        homcoord.Pt(rayon * np.cos(following_angle),
                                    rayon * np.sin(following_angle))
                    segments_list.append(homcoord.Segment(Pt1, Pt2))
            else:
                logging.warning('Unknown entity %s in dxf file' % e)

        return segments_list

    def shapes(self, interpolate_curve=None, ent=None):
        '''
        Translates drawing into list of shapely objects
        '''
        
        if interpolate_curve is None:
            interpolate_curve = self.interp

        shapes = []

        for e in self.entities():
            if e.dxftype == 'LINE':
                logging.warning(
                    'Open line in dxf file ignored; only closed shapes are '
                    'considered.')
            elif e.dxftype == 'CIRCLE':
                c = homcoord.Pt(e.center[:2])
                rayon = e.radius
                the_angles = np.linspace(-np.pi, np.pi, interpolate_curve)

                pts_polygon = []
                for angle in the_angles:
                    Pt1 = c + homcoord.Pt(rayon * np.cos(angle),
                                          rayon * np.sin(angle))
                    pts_polygon.append((Pt1.x, Pt1.y))

                shapes.append(Polygon(pts_polygon))
            elif e.dxftype == 'POLYLINE':
                pts_polygon = []
                for vertex in e.vertices:
                    point = homcoord.Pt(vertex.location[:2])
                    pts_polygon.append((point.x, point.y))
                shapes.append(Polygon(pts_polygon))
            elif e.dxftype == 'LWPOLYLINE':
                shapes.append(Polygon(e.points))
            else:
                logging.warning(
                    'Unknown entity {} ({}) in dxf shapes file'.format(
                        e, e.dxftype))

        return shapes


# ----- #
# Tools #
# ----- #

def Trans(scale=1, offset=[0, 0], rotation=0):
    __author__ = "Philippe Guglielmetti"
    __copyright__ = "Copyright 2013, Philippe Guglielmetti"
    __credits__ = ['http://effbot.org/imagingbook/imagedraw.htm']
    __license__ = "LGPL"
    res = homcoord.Xform(
        [[scale, 0, offset[0]], [0, scale, offset[1]], [0, 0, 1]])
    if rotation:
        res = homcoord.Xrotate(rotation * pi / 180.) * res
    return res


def plot_coords(ax, ob):
    x, y = ob.xy
    ax.plot(x, y, 'o', color='#999999', zorder=1)



#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# This file is part of the PyNCulture project, which aims at providing tools to
# easily generate complex neuronal cultures.
# Copyright (C) 2017 SENeC Initiative
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The tools present in this file are modified from the homcoord.py file by
# John W. Shipman (see __credits__)
#
# Contributor: S. Bottani

"""
2D Homogeneous coordinates with transformations

With addition of the SEGMENT class
S. Bottani
"""

__credits__ = ['http://www.nmt.edu/tcc/help/lang/python/examples/homcoord/']


import sys
from math import *

import numpy as np


# --------- #
# Constants #
# --------- #

RAD_45  = pi / 4.  # 45 degrees in radians
RAD_90  = pi / 2.  # 90 degrees in radians
RAD_180 = pi       # 180 degrees in radians
TWO_PI  = 2. * pi  # 360 degrees in radians


# ----------- #
# Point class #
# ----------- #

class Pt(object):

    '''
    Represents a homogeneous coordinate in 2-space.

    Exports:
    Pt(*coords):
      [ coords is a 2-sequence or a 1-sequence containing a
        2-sequence ->
          return a new Pt instance representing those two
          values as x and y, respectively
        coords is a 1-sequence containing a 3-sequence ->
          return a new Pt instance representing those values
          as x, y, and w, respectively ]
    .xy:
      [ return a 2-tuple with the homogenized x and y values ]
    .x:    [ return the homogenized x coordinate ]
    .y:    [ return the homogenized y coordinate ]
    .dist(other):
      [ other is a Pt instance ->
          return the distance between self and other ]
    .bearing(p):
      [ p is a Pt instance ->
          return the Cartesian angle in radians from self to p ]
    .radial(d, bearing):
      [ (d is a distance) and (bearing is an angle in radians) ->
          return the location at that distance and bearing as
          a Pt instance ]
    .toPolar():
      [ return self in polar coordinates as a Polar instance ]
    .__str__():  [ return self as a string ]
    .__add__(self, other):
      [ other is a Pt instance ->
          return a new Pt instance whose coordinates are the
          sum of self's and other's ]
    .__sub__(self, other):
      [ other is a Pt instance ->
          return a new Pt instance whose coordinates are the
          self's minus other's ]
    .__cmp__(self, other):
      [ if self and other are the same point ->
          return 0
        else -> return a nonzero value ]
    State/Invariants:
    .v     [ a numpy 3-element vector [x, y, W] ]
    '''

    def __init__ ( self, *args ):
        '''Constructor.
        '''
        #-- 1 --
        # [ if args has one element containing exactly two values ->
        #     x  :=  args[0][0]
        #     y  :=  args[0][1]
        #     w  :=  1.0
        #   else if args has one element containing three values ->
        #     x  :=  args[0][0]
        #     y  :=  args[0][1]
        #     w  :=  args[0][2]
        #   else if args has exactly two values ->
        #     x  :=  args[0]
        #     y  :=  args[1]
        #     w  :=  1.0
        #   else -> raise ValueError ]
        if len(args) == 1:
            value = args[0]
            if len(value) == 2:
                x, y = value
                w = 1.0
            else:
                x, y, w = value
        else:
            x, y = args
            w = 1.0
        #-- 2 --
        # [ self.v  :=  a 3-element numpy vector (x, y, w) as
        #                 type float
        #   self.__inverse  :=  None ]
        w=float(w) #to force floating division in x,y properties
        self.v = (x, y, w)

    @property
    def xy(self):
        '''
        Return (x,y)
        '''
        w = self.v[2]
        return (self.v[0]/w, self.v[1]/w)

    @property
    def x(self):
        '''
        Return the abscissa.
        '''
        return self.v[0]/self.v[2]

    @property
    def y(self):
        '''Return the ordinate.
        '''
        return self.v[1]/self.v[2]

    def apply(self,f):
        """:return: Pt obtained by appying function f to x and y"""
        return Pt(f(self.x),f(self.y))

    def dist(self, other):
        '''Return the distance between self and other.
        '''
        (dx,dy) = (self-other).xy
        return sqrt ( dx*dx + dy*dy )

    def bearing(self, p):
        '''
        What is the bearing angle from self to p?
        '''
        return atan2(p.y-self.y, p.x-self.x)

    def radial(self, d, bearing):
        '''
        Return the point at a given distance and bearing.
        '''
        return Pt(self.x + d*cos(bearing),
                  self.y + d*sin(bearing) )

    def toPolar(self):
        '''Convert to polar coordinates.
        '''
        x, y = self.xy
        return Polar(sqrt(x*x + y*y), atan2(y, x))

    def __str__(self):
        '''Return a string representation of self.
        '''
        return "(%.4g, %.4g)" % self.xy

    def __repr__(self):
        return str(self)

    def __add__(self, other):
        '''Add two points.
        '''
        return Pt(self.x+other.x, self.y+other.y)

    def __sub__(self, other):
        '''Subtract two points.
        '''
        return Pt(self.x-other.x, self.y-other.y)

    def __mul__(self, scale):
        '''Multiply by scalar.
        '''
        return Pt((self.x, self.y, 1./scale))

    def __div__(self, scale):
        '''Multiply by scalar.
        '''
        return Pt((self.x, self.y, scale))

    def __cmp__(self, other):
        '''Compare two points.
        '''
        return cmp(self.xy, other.xy)


# -------------------- #
# Transformation class #
# -------------------- #

class Xform(object):

    '''
    Represents an arbitrary homogeneous coordinate transform.

    Exports:
    Xform(m):
      [ m is a 3x3 transform matrix as a array, or
        a sequence that array() will accept as a 3x3
        array ->
          return a new Xform instance representing that
          transform ]
    .apply(p):
      [ p is a Pt instance ->
          return a new Pt instance representing p transformed
          by self ]
    .invert(p):
      [ p is a Pt instance ->
          return a new Pt instance pp such that
          self.apply(pp) == p ]
    .inverse():
      [ return the inverse of self as an Xform instance ]
    .compose(t):
      [ t is an Xform instance ->
          return a new Xform representing the composition of
          self followed by t ]
    .offset():
      [ return the net offset that self will shift the origin,
        as a Pt instance ]
    .angle():
      [ return the net angle that self will rotate the unit
        vector from (0,0) to (1,1) ]
    .mag():
      [ return the net magnification that self will apply to the
        unit vector ]
    .__str__(self):
      [ return a string representation of self ]
    State/Invariants:
    self._m:
      [ a 3x3 array representing the argument passed
        to the constructor ]
    self._mInverse:
      [ the inverse of self._m or None ]
    self.__offset:
      [ the net translation of self or None ]
    self.__angle:
      [ the net rotation of self or None ]
    self._mag:
      [ the net uniform scaling of self or None ]
    ORIGIN:      [ the origin as a Pt instance ]
    UNIT:        [ a point 1.0 along the line x=y ]
    '''

    ORIGIN = Pt(0,0)
    UNIT = ORIGIN.radial(1.0, RAD_45)

    def __init__ ( self, m ):
        '''Constructor.
        '''
        #-- 1 --
        # [ if the type of m is ndarray ->
        #     self._m  :=  m
        #   else if m is acceptable as an argument to
        #   array() ->
        #     self._m  :=  array(m)
        #   else -> raise Exception ]
        self._m = m
        #-- 2 --
        self._mInverse = None

    def apply ( self, p ):
        '''Transform a point.
        '''
        #-- 1 --
        # [ pp  :=  a array representing the dot product of
        #           self._m and p.v ]
        pp = dot(self._m, p.v)

        #-- 2 --
        # [ return a Pt instance representing pp.v ]
        return Pt(pp)

    def __call__(self, p):
        return self.apply(p)

    def invert ( self, p ):
        '''Return p transformed by the inverse of self, as a Pt.
        '''
        return self.inverse().apply ( p )

    def inverse ( self ):
        '''Return the inverse transform as an Xform.
        '''
        #-- 1 --
        # [ if self._mInverse is None ->
        #     self._mInverse  :=  matrix inverse of self._m
        #   else -> I ]
        if self._mInverse is None:
            self._mInverse = linalg.inv ( self._m )

        #-- 2 --
        return Xform(self._mInverse)

    def __str__(self):
        '''Display self as a string
        '''
        #-- 1 --
        return ( "<Xform(xlate(%s), rotate(%.1fdeg), "
                 "mag(%.1f)>" %
                 (self.offset(), degrees(self.angle()), self.mag()) )

    def compose ( self, t2 ):
        '''Return the composition of two transforms.
        '''
        return Xform ( dot ( t2._m, self._m ) )

    def __mul__ ( self, other ):
        '''Implement '*'
        '''
        return self.compose(other)

    def offset(self):
        return self(self.ORIGIN )

    def angle(self,angle=RAD_45):
        """
        :param angle: angle in radians of a unit vector starting at origin
        :return: float bearing in radians of the transformed vector
        """
        pt=Polar(1.0,angle).toCartesian()
        pt=self(pt)-self.offset()
        return atan2(pt.y,pt.x)


    def mag(self):
        '''Return the net (uniform) scaling of this transform.
        '''
        return self(self.ORIGIN ).dist(self(self.UNIT))


# ----------- #
# Polar class #
# ----------- #

class Polar(object):
    '''
    Represents a point in polar coordinates.

    Exports:
    Polar(r, theta):
      [ r and theta are numbers ->
          return a new Polar instance representing radius r
          and angle theta ]
    .r, .theta:  [ as passed to constructor ]
    .toCartesian():
      [ return self in Cartesian coordinates as a Pt instance ]
    .__str__():
      [ return self as a string "(r, theta)" ]
    '''

    def __init__(self, *p):
        '''Constructor
        '''
        self.r, self.theta = argPair(*p)

    def toCartesian(self):
        '''Return self in rectangular coordinates as a Pt instance.
        '''
        return Pt(self.r * cos(self.theta),
                  self.r * sin(self.theta))

    def __str__(self):
        '''Return self as a string.
        '''
        return ( "(%.4g, %.4gd)" %
                 (self.r, degrees(self.theta)) )


# ---------- #
# Line class #
# ---------- #

class Line(object):

    '''
    Represents a geometric line.

    Exports:
    Line(a, b, c):
      [ a, b, and c are floats ->
          return a Line instance representing ax+by+c=0 ]
    .a, .b, .c:  [ as passed to constructor, read-only ]
    .__str__(self):   [ return self as a string ]
    .intersect(other):
      [ other is a Line instance ->
          if self and other intersect ->
            return the intersection as a Pt
          else -> raise ValueError ]
    Line.twoPoint(p1, p2):       # Static method
      [ p1 and p2 are Pt instances ->
          if p1 and p2 are distinct ->
            return a Line instance representing the line that
            intersects p1 and p2
          else -> raise ValueError ]
    Line.pointBearing(p, bears):   # Static method
      [ (p is a Pt instance) and
        (bears is a Cartesian bearing in radians) ->
          return the line through p at bearing (bears) ]
    '''

    def __init__(self, a, b, c):
        '''Constructor.
        '''
        self.a = float(a)
        self.b = float(b)
        self.c = float(c)

    def __str__(self):
        '''Return a string representing self.
        '''
        return "%.4gx + %.4gy + %.4g = 0" % (self.a, self.b, self.c)

    def intersect(self, other):
        '''Where do lines self and other intersect?
        '''
        #-- 1 --
        # [ if self and other have the same slope ->
        #     raise ValueError
        #   else -> I ]
        if self.a * other.b == other.a * self.b:
            raise ValueError("Lines have the same slope.")
        #-- 2 --
        # [ x, y  :=  solution to the simultaneous linear equations
        #       (self.a * x + self.b * y = -self.c) and
        #       (other.a * x + other.b * y = -other.c) ]
        a = array ( ( (self.a, self.b), (other.a, other.b) ) )
        b = array ( (-self.c, -other.c) )
        x, y = linalg.solve(a,b)

        #-- 3 --
        return Pt(x, y)

    @staticmethod
    def twoPoint(p1, p2):
        '''Find the equation of a line between two points.
        '''
        #-- 1 --
        # [ if p1 and p2 coincide ->
        #     raise ValueError
        #   else ->
        #     x1  :=  abscissa of p1
        #     y1  :=  ordinate of p1
        #     x2  :=  abscissa of p2
        #     y2  :=  ordinate of p2 ]
        if p1 == p2:
            raise ValueError("Points are not distinct.")
        else:
            x1, y1 = p1.xy
            x2, y2 = p2.xy
        #-- 2 --
        # [ if x1 == x2 ->
        #     return a vertical line through x1
        #   else ->
        #     m  :=  (y2-y1)/(x2-x1) ]
        if x1 == x2:
            return Line(1.0, 0.0, -x1)
        else:
            m = (y2-y1)/(x2-x1)
        #-- 4 --
        # [ return a new Line instance having a=(-m), b=1, and
        #   c=(m*x1-y1) ]
        return Line(-m, 1.0, (m*x1-y1))

    @staticmethod
    def pointBearing(p, bears):
        '''Line through p at angle (bears).
        '''
        #-- 1 --
        # [ angle  :=  angle normalized to [0,180) degrees
        #   px  :=  abscissa of p
        #   py  :=  ordinate of p ]
        angle = bears % RAD_180
        px, py = p.xy

        #-- 2 --
        # [ if angle == RAD_90 ->
        #     return a Line with a=1.0, b=0.0, and c=-p.x
        #   else ->
        #     m  :=  tan(angle) ]
        if angle == RAD_90:
            return Line(1.0, 0.0, -px)
        else:
            m = tan(angle)
        #-- 3 --
        # [ return a Line with a=m, b=-1.0, and c=(-m*px + py) ]
        return Line(m, -1.0, py - m*px)

def argPair(*p):
    '''Process a pair of values passed in various ways.

      [ if len(p) is 2 ->
            return (p[0], p[1])
        else if p is a single non-iterable ->
          return (p[0], p[0])
        else if p is an iterable with two values ->
            return (p[0][0], p[0][1])
        else if p is an iterable with one value ->
            return (p[0][0], p[0][0])
    '''
    #-- 1 --
    if len(p) == 2:
        return (p[0], p[1])

    #-- 2 --
    it = p[0]
    if not hasattr(it, "__iter__"):
        return(it, it)

    #-- 3 --
    # [ p is an iterable ->
    #     values  :=  all values from p[0] ]
    values = [ x
               for x in p[0] ]

    #-- 4 --
    if len(values) == 1:
        return (values[0], values[0])
    return (values[0], values[1])


# ------------- #
# Segment class #
# ------------- #

class Segment(object):
    '''Represents a geometric Segment.

      Exports:
        Segment(a, b):
          [ a, b, are Points ->
              returns a Segment instance representing the segment between points a and b
              .length is the segment length
              .dir the unit vector in the direction a to b]

        .__str__(self):   [ return self as a string ]
        .intersect(other):
          [ other is a Segment instance ->
              if self and other intersect ->
                return the intersection as a Pt
              else -> raise ValueError ]

    '''

    def __init__(self, a, b):
        '''Constructor.
        '''
        self.a = a # a, and b Pt
        self.b = b
        self.length=np.sqrt((b.x-a.x)**2+(b.y-a.y)**2)
        self.dir=np.array((b.x-a.x),(b.y-a.y))/self.length

    def __str__(self):
        '''Return a string representing self.
        '''
        return "(%.4gx , %.4gy )" % (self.a, self.b)

    def intersect(self, other):
        ''' If segments self and other intersect return the intersection point as
        a Pt instance
        '''
        #-- 1 --
        # [ if self and other have the same slope ->
        #     raise ValueError
        #   else -> I ]
        if self.dir == other.dir:
            return
        #    raise ValueError("Lines have the same slope.")
        #-- 2 --
        # [ x, y  :=  solution to the simultaneous linear equations
        #       (self.a * x + self.b * y = -self.c) and
        #       (other.a * x + other.b * y = -other.c) ]


        line1=Line.twoPoints(self.a,self.b)
        line2=Line.twoPoints(other.a,other.b)

        intersection=line1.intersect(line2) # returns 0 it the 2 lines do not intersect

        # Check if intersection is in the segments

        self_in_segment = is_in_segment(intersection, self)
        other_in_segment = is_in_segment(intersection,other)
        if self_in_segment and other_in_segment:
            return intersection
        return False


# ----- #
# Tools #
# ----- #

def dot(a,b):
    '''
    Generalized dot product.
    '''
    try: #vector*vector
        return sum(map( operator.mul, a, b))
    except:
        pass
    try: #matrix*vector
        return [dot(line,b) for line in a]
    except:
        pass
    #matrix*matrix
    res=[dot(a,col) for col in zip(*b)]
    return map(list, zip(*res))


def normAngle(theta):
    '''
    Normalize an angle in radians to [0, 2*pi).
    '''
    return theta % TWO_PI


def is_in_segment(pt,seg):
    '''
    Returns True if a point is in the segment.
    '''

    scaling_x=np.abs((pt.x-seg.b.x)/(pt.x-seg.a.x))
    scaling_y=np.abs((pt.y-seg.b.y)/(pt.y-seg.a.y))

    if scaling_x != scaling_y :
        return False

    if scaling_x >1. :
        return False

    return True


# TRANSFORM FUNCTIONS

def Xlate(*p):
    '''
    Create a translation transform.
    '''
    #-- 1 --
    # [ dx  :=  first value from p
    #   dy  :=  second value from p ]
    dx, dy = argPair ( *p )

    #-- 2 --
    return Xform ( [ (1, 0, dx),
                     (0, 1, dy),
                     (0, 0, 1)  ] )


def Xscale(*p):
    '''
    Create a scaling transform.
    '''
    #-- 1 --
    # [ if p is a single value or single-valued iterable ->
    #     sx  :=  that value
    #     sy  :=  that value
    #   else ->
    #     sx  :=  the first value from p
    #     sy  :=  the second value from p ]
    sx, sy = argPair ( *p )

    #-- 2 --
    # [ return an Xform for scaling x by sx and scaling y by sy ]
    return Xform ( [ (sx, 0,  0),
                     (0,  sy, 0),
                     (0,  0,  1) ] )


def Xrotate(theta):
    '''Create a rotation transform.
    '''
    #-- 1 --
    sint = sin(theta)
    cost = cos(theta)

    #-- 2 --
    return Xform ( [ (cost, -sint, 0),
                     (sint, cost,  0),
                     (0,    0,     1) ] )


def Xrotaround ( p, theta ):
    '''Rotation of theta radians around point p.
    '''
    #-- 1 --
    # [ t1  :=  an Xform that translates point p to the origin
    #   r  :=  an Xform that rotates theta radians around the origin
    #   t2  :=  an Xform that translates the origin to point p ]
    t1 = Xlate ( [ -v
                   for v in p.xy ] )
    r = Xrotate ( theta )
    t2 = Xlate ( p.xy )

    #-- 2 --
    # [ return an Xform instance representing t1, then r, then t2 ]
    return t1.compose(r).compose(t2)



#-*- coding:utf-8 -*-
#
# core/nngt_graph.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Default (limited) graph if none of the graph libraries are available """

from collections import OrderedDict
from copy import deepcopy
import logging

import numpy as np
from scipy.sparse import coo_matrix, lil_matrix

import nngt
from nngt.lib import InvalidArgument, nonstring_container, is_integer
from nngt.lib.connect_tools import (_cleanup_edges, _set_dist_new_edges,
                                    _set_default_edge_attributes)
from nngt.lib.graph_helpers import _get_edge_attr, _get_dtype, _post_del_update
from nngt.lib.converters import _np_dtype, _to_np_array
from nngt.lib.logger import _log_message
from .graph_interface import GraphInterface, BaseProperty


logger = logging.getLogger(__name__)


# ---------- #
# Properties #
# ---------- #

class _NProperty(BaseProperty):

    ''' Class for generic interactions with nodes properties (graph-tool)  '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.prop = OrderedDict()

    def __getitem__(self, name):
        dtype = _np_dtype(super().__getitem__(name))
        return _to_np_array(self.prop[name], dtype=dtype)

    def __setitem__(self, name, value):
        if name in self:
            size = self.parent().node_nb()
            if len(value) == size:
                self.prop[name] = list(value)
            else:
                raise ValueError("A list or a np.array with one entry per "
                                 "node in the graph is required")
        else:
            raise InvalidArgument("Attribute does not exist yet, use "
                                  "set_attribute to create it.")

    def new_attribute(self, name, value_type, values=None, val=None):
        dtype = object
        if val is None:
            if value_type == "int":
                val = int(0)
                dtype = int
            elif value_type == "double":
                val = np.NaN
                dtype = float
            elif value_type == "string":
                val = ""
            else:
                val = None
                value_type = "object"

        if values is None:
            values = _to_np_array(
                [deepcopy(val) for _ in range(self.parent().node_nb())],
                value_type)

        if len(values) != self.parent().node_nb():
            raise ValueError("A list or a np.array with one entry per "
                             "node in the graph is required")

        # store name and value type in the dict
        super().__setitem__(name, value_type)

        # store the real values in the attribute
        self.prop[name] = list(values)
        self._num_values_set[name] = len(values)

    def set_attribute(self, name, values, nodes=None):
        '''
        Set the node attribute.

        Parameters
        ----------
        name : str
            Name of the node attribute.
        values : array, size N
            Values that should be set.
        nodes : array-like, optional (default: all nodes)
            Nodes for which the value of the property should be set. If `nodes`
            is not None, it must be an array of size N.
        '''
        num_nodes = self.parent().node_nb()
        num_n = len(nodes) if nodes is not None else num_nodes
        if num_n == num_nodes:
            self[name] = list(values)
            self._num_values_set[name] = num_nodes
        else:
            if num_n != len(values):
                raise ValueError("`nodes` and `values` must have the same "
                                 "size; got respectively " + str(num_n) + \
                                 " and " + str(len(values)) + " entries.")

            if self._num_values_set[name] == num_nodes - num_n:
                self.prop[name].extend(values)
            else:
                for n, val in zip(nodes, values):
                    self.prop[name][n] = val

        self._num_values_set[name] = num_nodes

    def remove(self, nodes):
        ''' Remove entries for a set of nodes '''
        for key in self:
            for n in reversed(sorted(nodes)):
                self.prop[key].pop(n)

            self._num_values_set[key] -= len(nodes)


class _EProperty(BaseProperty):

    ''' Class for generic interactions with nodes properties (graph-tool)  '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.prop = OrderedDict()

    def __getitem__(self, name):
        '''
        Return the attributes of an edge or a list of edges.
        '''
        eprop = {}
        graph = self.parent()

        if isinstance(name, slice):
            for k in self.keys():
                dtype = _np_dtype(super().__getitem__(k))
                eprop[k] = _to_np_array(self.prop[k], dtype)[name]

            return eprop
        elif nonstring_container(name):
            if nonstring_container(name[0]):
                eids = [graph.edge_id(e) for e in name]

                for k in self.keys():
                    dtype = _np_dtype(super().__getitem__(k))
                    eprop[k] = _to_np_array(self.prop[k], dtype=dtype)[eids]
            else:
                eid = graph.edge_id(name)

                for k in self.keys():
                    eprop[k] = self.prop[k][eid]

            return eprop
            
        dtype = _np_dtype(super().__getitem__(name))

        return _to_np_array(self.prop[name], dtype=dtype)

    def __setitem__(self, name, value):
        if name in self:
            size = self.parent().edge_nb()
            if len(value) == size:
                self.prop[name] = list(value)
            else:
                raise ValueError("A list or a np.array with one entry per "
                                 "edge in the graph is required")
        else:
            raise InvalidArgument("Attribute does not exist yet, use "
                                  "set_attribute to create it.")
        self._num_values_set[name] = len(value)

    def set_attribute(self, name, values, edges=None, last_edges=False):
        '''
        Set the edge property.

        Parameters
        ----------
        name : str
            Name of the edge property.
        values : array
            Values that should be set.
        edges : array-like, optional (default: None)
            Edges for which the value of the property should be set. If `edges`
            is not None, it must be an array of shape `(len(values), 2)`.
        '''
        num_edges = self.parent().edge_nb()
        num_e     = len(edges) if edges is not None else num_edges

        if num_e != len(values):
            raise ValueError("`edges` and `values` must have the same "
                             "size; got respectively " + str(num_e) + \
                             " and " + str(len(values)) + " entries.")

        if edges is None:
            self[name] = list(values)
        else:
            if last_edges:
                self.prop[name].extend(values)
            else:
                eid  = self.parent().edge_id
                prop = self.prop[name]

                # using list comprehension for fast loop
                [_set_prop(prop, eid(e), val) for e, val in zip(edges, values)]

        if num_e:
            self._num_values_set[name] = num_edges

    def new_attribute(self, name, value_type, values=None, val=None):
        num_edges = self.parent().edge_nb()

        if values is None and val is None:
            self._num_values_set[name] = num_edges

        if val is None:
            if value_type == "int":
                val = int(0)
            elif value_type == "double":
                val = np.NaN
            elif value_type == "string":
                val = ""
            else:
                val = None

        if values is None:
            values = _to_np_array(
                [deepcopy(val) for _ in range(num_edges)],
                value_type)

        if len(values) != num_edges:
            self._num_values_set[name] = 0
            raise ValueError("A list or a np.array with one entry per "
                             "edge in the graph is required")

        # store name and value type in the dict
        super().__setitem__(name, value_type)

        # store the real values in the attribute
        self.prop[name] = list(values)
        self._num_values_set[name] = len(values)

    def edges_deleted(self, eids):
        ''' Remove the attributes of a set of edge ids '''
        for key, val in self.prop.items():
            self.prop[key] = [v for i, v in enumerate(val) if i not in eids]
            self._num_values_set[key] -= len(eids)


# ----------------- #
# NNGT backup graph #
# ----------------- #

class _NNGTGraphObject:
    '''
    Minimal implementation of the GraphObject, which does not rely on any
    graph-library.
    '''

    def __init__(self, nodes=0, weighted=True, directed=True):
        ''' Initialized independent graph '''
        self._nodes    = set(i for i in range(nodes))
        self._out_deg  = [0]*nodes
        self._in_deg   = [0]*nodes

        if directed:
            # for directed networks, edges and unique are the same
            self._edges = self._unique = OrderedDict()
            assert self._edges is self._unique
        else:
            # for undirected networks
            self._edges  = OrderedDict()
            self._unique = OrderedDict()

        self._directed = directed
        self._weighted = weighted

    def copy(self):
        ''' Returns a deep copy of the graph object '''
        copy = _NNGTGraphObject(len(self._nodes), weighted=self._weighted,
                                directed=self._directed)

        copy._nodes   = self._nodes.copy()

        if self._directed:
            copy._unique = copy._edges = self._edges.copy()
            assert copy._unique is copy._edges
        else:
            copy._edges  = self._edges.copy()
            copy._unique = self._unique.copy()

        copy._out_deg = self._out_deg.copy()
        copy._in_deg  = self._in_deg.copy()

        return copy

    def is_directed(self):
        return self._directed

    @property
    def nodes(self):
        return list(self._nodes)


class _NNGTGraph(GraphInterface):
    ''' NNGT wrapper class for _NNGTGraphObject '''

    #------------------------------------------------------------------#
    # Constructor and instance properties

    def __init__(self, nodes=0, weighted=True, directed=True,
                 copy_graph=None, **kwargs):
        ''' Initialized independent graph '''
        self._nattr = _NProperty(self)
        self._eattr = _EProperty(self)

        self._max_eid = 0

        # test if copying graph
        if copy_graph is not None:
            self._from_library_graph(copy_graph, copy=True)
            self._max_eid = copy_graph._max_eid
        else:
            self._graph = _NNGTGraphObject(
                nodes=nodes, weighted=weighted, directed=directed)

    #------------------------------------------------------------------#
    # Graph manipulation

    def edge_id(self, edge):
        '''
        Return the ID a given edge or a list of edges in the graph.
        Raises an error if the edge is not in the graph or if one of the
        vertices in the edge is nonexistent.

        Parameters
        ----------
        edge : 2-tuple or array of edges
            Edge descriptor (source, target).

        Returns
        -------
        index : int or array of ints
            Index of the given `edge`.
        '''
        g = self._graph

        if is_integer(edge[0]):
            return g._edges[tuple(edge)]
        elif nonstring_container(edge[0]):
            idx = [g._edges[tuple(e)] for e in edge]
            return idx
        else:
            raise AttributeError("`edge` must be either a 2-tuple of ints or "
                                 "an array of 2-tuples of ints.")

    def has_edge(self, edge):
        '''
        Whether `edge` is present in the graph.

        .. versionadded:: 2.0
        '''
        e = tuple(edge)

        return e in self._graph._edges

    @property
    def edges_array(self):
        '''
        Edges of the graph, sorted by order of creation, as an array of
        2-tuple.
        '''
        return np.asarray(list(self._graph._unique), dtype=int)

    def _get_edges(self, source_node=None, target_node=None):
        '''
        Called by Graph.get_edges if source_node and target_node are not both
        integers.
        '''
        g = self._graph

        if source_node is not None:
            source_node = \
                {source_node} if is_integer(source_node) else set(source_node)

            # source only
            if target_node is None:
                if g.is_directed():
                    return [e for e in g._unique if e[0] in source_node]

                return [e for e in g._unique
                        if e[0] in source_node or e[1] in source_node]

            # source and target
            target_node = \
                {target_node} if is_integer(target_node) else set(target_node)

            if g.is_directed():
                return [e for e in g._unique
                        if e[0] in source_node and e[1] in target_node]

            return [e for e in g._unique
                    if e[0] in source_node and e[1] in target_node
                    or e[1] in source_node and e[0] in target_node]

        if target_node is None:
            # return all edges
            return list(g._unique)

        # target only
        target_node = \
            {target_node} if is_integer(target_node) else set(target_node)

        if g.is_directed():
            return [e for e in g._unique if e[1] in target_node]

        return [e for e in g._unique
                if e[0] in target_node or e[1] in target_node]

    def is_connected(self):
        raise NotImplementedError("Not available with 'nngt' backend, please "
                                  "install a graph library (networkx, igraph, "
                                  "or graph-tool).")

    def new_node(self, n=1, neuron_type=1, attributes=None, value_types=None,
                 positions=None, groups=None):
        '''
        Adding a node to the graph, with optional properties.

        Parameters
        ----------
        n : int, optional (default: 1)
            Number of nodes to add.
        neuron_type : int, optional (default: 1)
            Type of neuron (1 for excitatory, -1 for inhibitory)
        attributes : dict, optional (default: None)
            Dictionary containing the attributes of the nodes.
        value_types : dict, optional (default: None)
            Dict of the `attributes` types, necessary only if the `attributes`
            do not exist yet.
        positions : array of shape (n, 2), optional (default: None)
            Positions of the neurons. Valid only for
            :class:`~nngt.SpatialGraph` or :class:`~nngt.SpatialNetwork`.
        groups : str, int, or list, optional (default: None)
            :class:`~nngt.core.NeuralGroup` to which the neurons belong. Valid
            only for :class:`~nngt.Network` or :class:`~nngt.SpatialNetwork`.

        Returns
        -------
        The node or a tuple of the nodes created.
        '''
        nodes = []

        g = self._graph

        if n == 1:
            nodes.append(len(g._nodes))
            g._in_deg.append(0)
            g._out_deg.append(0)
        else:
            num_nodes = len(g._nodes)
            nodes.extend(
                [i for i in range(num_nodes, num_nodes + n)])
            g._in_deg.extend([0 for _ in range(n)])
            g._out_deg.extend([0 for _ in range(n)])

        g._nodes.update(nodes)

        attributes = {} if attributes is None else deepcopy(attributes)

        if attributes:
            for k, v in attributes.items():
                if k not in self._nattr:
                    self._nattr.new_attribute(k, value_types[k], val=v)
                else:
                    v = v if nonstring_container(v) else [v]
                    self._nattr.set_attribute(k, v, nodes=nodes)

        # set default values for all attributes that were not set
        for k in self.node_attributes:
            if k not in attributes:
                dtype = self.get_attribute_type(k)
                if dtype == "double":
                    values = [np.NaN for _ in nodes]
                    self._nattr.set_attribute(k, values, nodes=nodes)
                elif dtype == "int":
                    values = [0 for _ in nodes]
                    self._nattr.set_attribute(k, values, nodes=nodes)
                elif dtype == "string":
                    values = ["" for _ in nodes]
                    self._nattr.set_attribute(k, values, nodes=nodes)
                else:
                    values = [None for _ in nodes]
                    self._nattr.set_attribute(k, values, nodes=nodes)

        if self.is_spatial():
            old_pos      = self._pos
            self._pos    = np.full((self.node_nb(), 2), np.NaN, dtype=float)
            num_existing = len(old_pos) if old_pos is not None else 0
            if num_existing != 0:
                self._pos[:num_existing, :] = old_pos
        if positions is not None:
            assert self.is_spatial(), \
                "`positions` argument requires a SpatialGraph/SpatialNetwork."
            self._pos[nodes] = positions

        if groups is not None:
            assert self.is_network(), \
                "`positions` argument requires a Network/SpatialNetwork."
            if nonstring_container(groups):
                assert len(groups) == n, "One group per neuron required."
                for g, node in zip(groups, nodes):
                    self.population.add_to_group(g, node)
            else:
                self.population.add_to_group(groups, nodes)

        if n == 1:
            return nodes[0]

        return nodes

    def delete_nodes(self, nodes):
        '''
        Remove nodes (and associated edges) from the graph.
        '''
        g = self._graph

        old_nodes = range(self.node_nb())

        # update node set and degrees
        if nonstring_container(nodes):
            nodes = set(nodes)

            g._nodes = g._nodes.difference(nodes)
        else:
            g._nodes.remove(nodes)
            g._out_deg.pop(nodes)
            g._in_deg.pop(nodes)

            nodes = {nodes}

        # remove node attributes
        self._nattr.remove(nodes)

        # map from old nodes to new node ids
        idx = 0

        remapping = {}

        for n in old_nodes:
            if n not in nodes:
                remapping[n] = idx
                idx += 1

        # remove edges and remap edges
        remove_eids = set()

        new_edges  = OrderedDict()

        new_eid = 0

        for e, eid in g._unique.items():
            if e[0] in nodes or e[1] in nodes:
                remove_eids.add(eid)
            else:
                new_edges[(remapping[e[0]], remapping[e[1]])] = new_eid
                new_eid += 1

        g._unique = new_edges

        if not g._directed:
            g._edges = new_edges.copy()
            g._edges.update({e[::-1]: i for e, i in new_edges.items()})
        else:
            g._edges = g._unique

        # tell edge attributes
        self._eattr.edges_deleted(remove_eids)

        g._nodes = set(range(len(g._nodes)))

        # check spatial and structure properties
        _post_del_update(self, nodes, remapping=remapping)

    def new_edge(self, source, target, attributes=None, ignore=False,
                 self_loop=False):
        '''
        Adding a connection to the graph, with optional properties.

        .. versionchanged :: 2.0
            Added `self_loop` argument to enable adding self-loops.

        Parameters
        ----------
        source : :class:`int/node`
            Source node.
        target : :class:`int/node`
            Target node.
        attributes : :class:`dict`, optional (default: ``{}``)
            Dictionary containing optional edge properties. If the graph is
            weighted, defaults to ``{"weight": 1.}``, the unit weight for the
            connection (synaptic strength in NEST).
        ignore : bool, optional (default: False)
            If set to True, ignore attempts to add an existing edge and accept
            self-loops; otherwise an error is raised.
        self_loop : bool, optional (default: False)
            Whether to allow self-loops or not.

        Returns
        -------
        The new connection or None if nothing was added.
        '''
        g = self._graph

        attributes = {} if attributes is None else deepcopy(attributes)

        # set default values for attributes that were not passed
        _set_default_edge_attributes(self, attributes, num_edges=1)

        # check that the edge does not already exist
        edge = (source, target)

        if source not in g._nodes:
            raise InvalidArgument("There is no node {}.".format(source))

        if target not in g._nodes:
            raise InvalidArgument("There is no node {}.".format(target))

        if source == target:
            if not ignore and not self_loop:
                raise InvalidArgument("Trying to add a self-loop.")
            elif ignore:
                _log_message(logger, "INFO",
                             "Self-loop on {} ignored.".format(source))

                return None

        if (g._directed and edge not in g._unique) or edge not in g._edges:
            edge_id             = self._max_eid
            # ~ edge_id             = len(g._unique)
            g._unique[edge]     = edge_id
            g._out_deg[source] += 1
            g._in_deg[target]  += 1

            self._max_eid += 1

            # check distance
            _set_dist_new_edges(attributes, self, [edge])

            # attributes
            self._attr_new_edges([(source, target)], attributes=attributes)

            if not g._directed:
                # edges and unique are different objects, so update _edges
                g._edges[edge] = edge_id
                # add reciprocal
                e_recip             = (target, source)
                g._edges[e_recip]   = edge_id
                g._out_deg[target] += 1
                g._in_deg[source]  += 1
        else:
            if not ignore:
                raise InvalidArgument("Trying to add existing edge.")

            _log_message(logger, "INFO",
                         "Existing edge {} ignored.".format((source, target)))

        return edge

    def new_edges(self, edge_list, attributes=None, check_duplicates=False,
                  check_self_loops=True, check_existing=True,
                  ignore_invalid=False):
        '''
        Add a list of edges to the graph.

        .. versionchanged:: 2.0
            Can perform all possible checks before adding new edges via the
            ``check_duplicates`` ``check_self_loops``, and ``check_existing``
            arguments.

        Parameters
        ----------
        edge_list : list of 2-tuples or np.array of shape (edge_nb, 2)
            List of the edges that should be added as tuples (source, target)
        attributes : :class:`dict`, optional (default: ``{}``)
            Dictionary containing optional edge properties. If the graph is
            weighted, defaults to ``{"weight": ones}``, where ``ones`` is an
            array the same length as the `edge_list` containing a unit weight
            for each connection (synaptic strength in NEST).
        check_duplicates : bool, optional (default: False)
            Check for duplicate edges within `edge_list`.
        check_self_loops : bool, optional (default: True)
            Check for self-loops.
        check_existing : bool, optional (default: True)
            Check whether some of the edges in `edge_list` already exist in the
            graph or exist multiple times in `edge_list` (also performs
            `check_duplicates`).
        ignore_invalid : bool, optional (default: False)
            Ignore invalid edges: they are not added to the graph and are
            silently dropped. Unless this is set to true, an error is raised
            whenever one of the three checks fails.

        .. warning::

            Setting `check_existing` to False will lead to undefined behavior
            if existing edges are provided! Only use it (for speedup) if you
            are sure that you are indeed only adding new edges.

        Returns
        -------
        Returns new edges only.
        '''
        attributes = {} if attributes is None else deepcopy(attributes)
        num_edges  = len(edge_list)

        g = self._graph

        # set default values for attributes that were not passed
        _set_default_edge_attributes(self, attributes, num_edges)

        # check that all nodes exist
        if num_edges:
            if np.max(edge_list) >= self.node_nb():
                raise InvalidArgument("Some nodes do no exist.")

            # check edges
            new_attr = None

            if check_duplicates or check_self_loops or check_existing:
                edge_list, new_attr = _cleanup_edges(
                    self, edge_list, attributes, check_duplicates,
                    check_self_loops, check_existing, ignore_invalid)
            else:
                new_attr = attributes

            # create the edges
            initial_eid = self._max_eid

            ws        = None
            num_added = len(edge_list)

            if "weight" in new_attr:
                if nonstring_container(new_attr["weight"]):
                    ws = new_attr["weight"]
                else:
                    ws = (new_attr["weight"] for _ in range(num_added))
            else:
                ws = _get_edge_attr(self, edge_list, "weight", last_edges=True)

            for i, (e, w) in enumerate(zip(edge_list, ws)):
                eid = self._max_eid

                g._unique[tuple(e)] = eid

                self._max_eid += 1

                g._out_deg[e[0]] += 1
                g._in_deg[e[1]]  += 1

                if not g._directed:
                    # edges and unique are different objects, so update _edges
                    g._edges[tuple(e)] = eid
                    # reciprocal edge
                    g._edges[tuple(e[::-1])] = eid

                    g._out_deg[e[1]] += 1
                    g._in_deg[e[0]]  += 1

            # check distance
            _set_dist_new_edges(new_attr, self, edge_list)

            # call parent function to set the attributes
            self._attr_new_edges(edge_list, attributes=new_attr)

        return edge_list

    def delete_edges(self, edges):
        ''' Remove a list of edges '''
        g = self._graph

        old_enum = len(g._unique)

        if not nonstring_container(edges[0]):
            edges = [edges]

        if not isinstance(edges[0], tuple):
            edges = [tuple(e) for e in edges]

        # get edge ids
        e_to_eid = g._unique

        eids = {e_to_eid[e] for e in edges}

        # remove
        directed = g._directed

        for e in edges:
            if e in g._unique:
                del g._unique[e]

            if not directed:
                if e in g._edges:
                    del g._edges[e]
                    del g._edges[e[::-1]]

        # reset eids
        for i, e in enumerate(g._unique):
            g._unique[e] = i

            if not directed:
                e._edges[e] = i
                e._edges[e[::-1]] = i

        self._eattr.edges_deleted(eids)

    def clear_all_edges(self):
        g = self._graph

        if g._directed:
            g._edges = g._unique = OrderedDict()
            assert g._edges is g._unique
        else:
            g._edges  = OrderedDict()
            g._unique = OrderedDict()

        g._out_deg = [0 for _ in range(self.node_nb())]
        g._out_deg = [0 for _ in range(self.node_nb())]

        self._eattr.clear()

    #------------------------------------------------------------------#
    # Getters

    def node_nb(self):
        '''
        Returns the number of nodes.

        .. warning:: When using MPI, returns only the local number of nodes.
        '''
        return len(self._graph._nodes)

    def edge_nb(self):
        '''
        Returns the number of edges.

        .. warning:: When using MPI, returns only the local number of edges.
        '''
        return len(self._graph._unique)

    def is_directed(self):
        return g._directed

    def get_degrees(self, mode="total", nodes=None, weights=None):
        '''
        Returns the degree of the nodes.

        .. warning ::
            When using MPI, returns only the degree related to local edges.
        '''
        g = self._graph

        num_nodes = None
        weights   = 'weight' if weights is True else weights

        if nodes is None:
            num_nodes = self.node_nb()
            nodes = slice(num_nodes)
        elif nonstring_container(nodes):
            nodes = list(nodes)
            num_nodes = len(nodes)
        else:
            nodes = [nodes]
            num_nodes = 1

        # weighted
        if nonstring_container(weights) or weights in self._eattr:
            degrees = np.zeros(num_nodes)
            adj_mat = self.adjacency_matrix(types=False, weights=weights)

            if mode in ("in", "total") or not self.is_directed():
                degrees += adj_mat.sum(axis=0).A1[nodes]
            if mode in ("out", "total") and self.is_directed():
                degrees += adj_mat.sum(axis=1).A1[nodes]

            return degrees
        elif weights not in {None, False}:
            raise ValueError("Invalid `weights` {}".format(weights))

        # unweighted
        degrees = np.zeros(num_nodes, dtype=int)

        if not g._directed or mode in ("in", "total"):
            if isinstance(nodes, slice):
                degrees += g._in_deg[nodes]
            else:
                degrees += [g._in_deg[i] for i in nodes]

        if g._directed and mode in ("out", "total"):
            if isinstance(nodes, slice):
                degrees += g._out_deg[nodes]
            else:
                degrees += [g._out_deg[i] for i in nodes]

        if num_nodes == 1:
            return degrees[0]

        return degrees

    def neighbours(self, node, mode="all"):
        '''
        Return the neighbours of `node`.

        Parameters
        ----------
        node : int
            Index of the node of interest.
        mode : string, optional (default: "all")
            Type of neighbours that will be returned: "all" returns all the
            neighbours regardless of directionality, "in" returns the
            in-neighbours (also called predecessors) and "out" retruns the
            out-neighbours (or successors).

        Returns
        -------
        neighbours : set
            The neighbours of `node`.
        '''
        edges = self.edges_array

        if mode == "all" or not self._graph._directed:
            neighbours = set(edges[edges[:, 1] == node, 0])
            return neighbours.union(edges[edges[:, 0] == node, 1])

        if mode == "in":
            return set(edges[edges[:, 1] == node, 0])

        if mode == "out":
            return set(edges[edges[:, 0] == node, 1])

        raise ValueError(('Invalid `mode` argument {}; possible values'
                          'are "all", "out" or "in".').format(mode))

    def _from_library_graph(self, graph, copy=True):
        ''' Initialize `self._graph` from existing library object. '''
        self._graph = graph._graph.copy() if copy else graph._graph

        for key, val in graph._nattr.items():
            dtype = graph._nattr.value_type(key)
            self._nattr.new_attribute(key, dtype, values=val)

        for key, val in graph._eattr.items():
            dtype = graph._eattr.value_type(key)
            self._eattr.new_attribute(key, dtype, values=val)


# tool function to set edge properties

def _set_prop(array, eid, val):
    array[eid] = val



#-*- coding:utf-8 -*-
#
# io/graph_saving.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" IO tools for NNGT """

import codecs
import logging
import pickle
import sys
import weakref

from collections import defaultdict

import numpy as np
import scipy.sparse as ssp

import nngt
from nngt.lib import InvalidArgument, on_master_process
from nngt.lib.logger import _log_message

from ..geometry import Shape, _shapely_support
from .io_helpers import _get_format
from .saving_helpers import (_neighbour_list, _edge_list, _gml, _custom_info,
                           _gml_info, _str_bytes_len)


logger = logging.getLogger(__name__)


# ---------- #
# Formatting #
# ---------- #

di_format = {
    "neighbour": _neighbour_list,
    "edge_list": _edge_list,
    "gml": _gml
}

format_graph_info = defaultdict(lambda: _custom_info)
format_graph_info["gml"] = _gml_info


# --------------- #
# Saving function #
# --------------- #

def save_to_file(graph, filename, fmt="auto", separator=" ",
                 secondary=";", attributes=None, notifier="@"):
    '''
    Save a graph to file.

    @todo: implement dot, xml/graphml, and gt formats

    Parameters
    ----------
    graph : :class:`~nngt.Graph` or subclass
        Graph to save.
    filename: str
        The path to the file.
    fmt : str, optional (default: "auto")
        The format used to save the graph. Supported formats are: "neighbour"
        (neighbour list, default if format cannot be deduced automatically),
        "ssp" (scipy.sparse), "edge_list" (list of all the edges in the graph,
        one edge per line, represented by a ``source target``-pair), "gml"
        (gml format, default if `filename` ends with '.gml'), "graphml"
        (graphml format, default if `filename` ends with '.graphml' or '.xml'),
        "dot" (dot format, default if `filename` ends with '.dot'), "gt" (only
        when using `graph_tool <http://graph-tool.skewed.de/>`_ as library,
        detected if `filename` ends with '.gt').
    separator : str, optional (default " ")
        separator used to separate inputs in the case of custom formats (namely
        "neighbour" and "edge_list")
    secondary : str, optional (default: ";")
        Secondary separator used to separate attributes in the case of custom
        formats.
    attributes : list, optional (default: ``None``)
        List of names for the edge attributes present in the graph that will be
        saved to disk; by default (``None``), all attributes will be saved.
    notifier : str, optional (default: "@")
        Symbol specifying the following as meaningfull information. Relevant
        information are formatted ``@info_name=info_value``, with
        ``info_name`` in ("attributes", "attr_types", "directed", "name",
        "size").
        Additional notifiers are ``@type=SpatialGraph/Network/SpatialNetwork``,
        which are followed by the relevant notifiers among ``@shape``,
        ``@structure``, and ``@graph`` to separate the sections.

    Note
    ----
    Positions are saved as bytes by :func:`numpy.nparray.tostring`
    '''
    fmt = _get_format(fmt, filename)

    # check for mpi
    if nngt.get_config("mpi"):
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        size = comm.Get_size()
        rank = comm.Get_rank()
        # get the
        str_local, di_notif = _as_string(
            graph, separator=separator, fmt=fmt, secondary=secondary,
            attributes=attributes, notifier=notifier, return_info=True)
        # make notification string only on master thread
        str_notif = ""
        if on_master_process():
            for key, val in iter(di_notif.items()):
                str_notif += "{}{}={}\n".format(notifier, key, val)
        # strings need to start with a newline because MPI strips last
        str_local = "\n" + str_local
        # gather all strings sizes
        sizes = comm.allgather(
            _str_bytes_len(str_local) + _str_bytes_len(str_notif))
        # get rank-based offset
        offset = [_str_bytes_len(str_notif)]
        offset.extend(np.cumsum(sizes)[:-1])
        # open file and write
        if on_master_process():
            with open(filename, "w") as f_graph:
                f_graph.write(str_notif)
        # parallel write
        amode = MPI.MODE_WRONLY
        fh = MPI.File.Open(comm, filename, amode)
        fh.Write_at_all(offset[rank], str_local.encode('utf-8'))
        fh.Close()
    else:
        str_graph = _as_string(
            graph, separator=separator, fmt=fmt, secondary=secondary,
            attributes=attributes, notifier=notifier)
        with open(filename, "w") as f_graph:
            f_graph.write(str_graph)


# --------------------- #
# String representation #
# --------------------- #

def _as_string(graph, fmt="neighbour", separator=" ", secondary=";",
              attributes=None, notifier="@", return_info=False):
    '''
    Full string representation of the graph.

    Parameters
    ----------
    graph : :class:`~nngt.Graph` or subclass
        Graph to save.
    fmt : str, optional (default: "auto")
        The format used to save the graph. Supported formats are: "neighbour"
        (neighbour list, default if format cannot be deduced automatically),
        "ssp" (:mod:`scipy.sparse`), "edge_list" (list of all the edges in the
        graph, one edge per line, represented by a ``source target``-pair),
        "gml" (gml format, default if `filename` ends with '.gml'), "graphml"
        (graphml format, default if `filename` ends with '.graphml' or '.xml'),
        "dot" (dot format, default if `filename` ends with '.dot'), "gt" (only
        when using `graph_tool`<http://graph-tool.skewed.de/>_ as library,
        detected if `filename` ends with '.gt').
    separator : str, optional (default " ")
        separator used to separate inputs in the case of custom formats (namely
        "neighbour" and "edge_list")
    secondary : str, optional (default: ";")
        Secondary separator used to separate attributes in the case of custom
        formats.
    attributes : list, optional (default: all)
        List of names for the edge attributes present in the graph that will be
        saved to disk; by default, all attributes will be saved.
    notifier : str, optional (default: "@")
        Symbol specifying the following as meaningfull information. Relevant
        information are formatted ``@info_name=info_value``, with
        ``info_name`` in ("attributes", "attr_types", "directed", "name",
        "size").
        Additional notifiers are ``@type=SpatialGraph/Network/SpatialNetwork``,
        which are followed by the relevant notifiers among ``@shape``, ``@x``,
        ``@y``, ``@z``, ``@structure``, and ``@graph`` to separate the
        sections.

    Returns
    -------
    str_graph : string
        The full graph representation as a string.
    '''
    # checks
    if separator == secondary and fmt != "edge_list":
        raise InvalidArgument("`separator` and `secondary` strings must be "
                              "different.")

    if notifier == separator or notifier == secondary:
        raise InvalidArgument("`notifier` string should differ from "
                              "`separator` and `secondary`.")

    # temporarily disable numpy cut threshold to save string
    old_threshold = np.get_printoptions()['threshold']
    np.set_printoptions(threshold=sys.maxsize)

    # data
    if attributes is None:
        attributes = [a for a in graph.edge_attributes if a != "bweight"]

    nattributes = [a for a in graph.node_attributes]

    additional_notif = {
        "directed": graph.is_directed(),
        "node_attributes": nattributes,
        "node_attr_types": [
            graph.get_attribute_type(nattr, "node") for nattr in nattributes
        ],
        "edge_attributes": attributes,
        "edge_attr_types": [
            graph.get_attribute_type(attr, "edge") for attr in attributes
        ],
        "name": graph.name,
        "size": graph.node_nb()
    }

    # add node attributes to the notifications
    for nattr in additional_notif["node_attributes"]:
        key = "na_" + nattr

        tmp = np.array2string(
            graph.get_node_attributes(name=nattr), max_line_width=np.NaN,
            separator=separator)[1:-1].replace("'" + separator + "'",
                                               '"' + separator + '"')

        # replace possible variants
        tmp = tmp.replace("'" + separator + '"', '"' + separator + '"')
        tmp = tmp.replace('"' + separator + "'", '"' + separator + '"')

        if tmp.startswith("'"):
            tmp = '"' + tmp[1:]

        if tmp.endswith("'"):
             tmp = tmp[:-1] + '"'

        # make and store final string
        additional_notif[key] = tmp

    # save positions for SpatialGraph (and shape if Shapely is available)
    if graph.is_spatial():
        if _shapely_support:
            additional_notif['shape'] = graph.shape.wkt
            additional_notif['default_areas'] = \
                {k: v.wkt for k, v in graph.shape.default_areas.items()}
            additional_notif['default_areas_prop'] = \
                {k: v.properties for k, v in graph.shape.default_areas.items()}
            additional_notif['non_default_areas'] = \
                {k: v.wkt for k, v in graph.shape.non_default_areas.items()}
            additional_notif['non_default_areas_prop'] = \
                {k: v.properties
                 for k, v in graph.shape.non_default_areas.items()}
            additional_notif['unit'] = graph.shape.unit
            min_x, min_y, max_x, max_y = graph.shape.bounds
            additional_notif['min_x'] = min_x
            additional_notif['max_x'] = max_x
        else:
            _log_message(logger, "WARNING",
                         'The `shape` attribute of the graph could not be '
                         'saved to file because Shapely is not installed.')

        pos = graph.get_positions()
        additional_notif['x'] = np.array2string(
            pos[:, 0], max_line_width=np.NaN, separator=separator)[1:-1]
        additional_notif['y'] = np.array2string(
            pos[:, 1], max_line_width=np.NaN, separator=separator)[1:-1]
        if pos.shape[1] == 3:
            additional_notif['z'] = np.array2string(
                pos[:, 2], max_line_width=np.NaN, separator=separator)[1:-1]

    if graph.structure is not None:
        # temporarily remove weakrefs
        graph.structure._parent = None
        for g in graph.structure.values():
            g._struct = None
            g._net    = None
        # save as string
        if nngt.get_config("mpi"):
            if nngt.get_config("mpi_comm").Get_rank() == 0:
                additional_notif["structure"] = codecs.encode(
                    pickle.dumps(graph.structure, protocol=2),
                                 "base64").decode().replace('\n', '~')
        else:
            additional_notif["structure"] = codecs.encode(
                pickle.dumps(graph.structure, protocol=2),
                             "base64").decode().replace('\n', '~')
        # restore weakrefs
        graph.structure._parent = weakref.ref(graph)
        for g in graph.structure.values():
            g._struct = weakref.ref(graph.structure)
            g._net    = weakref.ref(graph)

    str_graph = di_format[fmt](graph, separator=separator,
                               secondary=secondary, attributes=attributes)

    # set numpy cut threshold back on
    np.set_printoptions(threshold=old_threshold)

    if return_info:
        return str_graph, additional_notif

    # format the info into the string
    info_str = format_graph_info[fmt](additional_notif, notifier, graph=graph)

    return info_str + str_graph



#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# base_test.py
#
# This file is part of the NNGT project to generate and analyze
# neuronal networks and their activity.
# Copyright (C) 2015-2019  Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Define the XmlHandler and the TestBasis class for NNGT testing module
"""

# std imports
import sys
import unittest
from abc import ABCMeta, abstractmethod, abstractproperty
from os import listdir
from os.path import abspath, dirname, isfile, join

# third-party library
import numpy as np
import xml.etree.ElementTree as xmlet

import nngt
from tools_testing import _bool_from_string, _xml_to_dict, _list_from_string


# ------------------- #
# Path to input files #
# ------------------- #

# folder containing all networks
test_dir = dirname(abspath(__file__)) + "/"
network_dir = test_dir + "Networks/"

# file containing networks' properties and instructions
xml_file = test_dir + "graph_tests.xml"


# --------------- #
# XmlResultParser #
# --------------- #

class XmlHandler:

    ''' Class parsing the XML reference files. '''

    di_type = {
        "float": float,
        "double": float,
        "string": str,
        "str": str,
        "int": int,
        "bool": _bool_from_string
    }

    def __init__(self):
        tree = xmlet.parse(xml_file)
        self.root = tree.getroot()
        self.graphs = self.root.find("graphs")

    def result(self, elt):
        return self.di_type[elt.tag](elt.text)

    def get_graph_list(self, test):
        elt_test = self.root.find('./test[@name="{}"]'.format(test))
        return [ child.text for child in elt_test.find("graph_list") ]

    def get_graph_options(self, graph):
        graph_elt = self.graphs.find('./graph[@name="{}"]'.format(graph))
        elt_options = None
        for child in graph_elt:
            if child.tag in ("load_options", "generate_options"):
                elt_options = child

        if elt_options:
            return _xml_to_dict(elt_options, self.di_type)

        return {}

    def get_result(self, graph, result_name):
        elt_test = self.graphs.find('./graph[@name="{}"]'.format(graph))
        elt_result = elt_test.find('./*[@name="{}"]'.format(result_name))
        return self.result(elt_result)


# --------------- #
# TestBasis class #
# --------------- #

class TestBasis(unittest.TestCase):

    '''
    Class defining the graphs and the conditions in which they will be tested.

    warning ::
        All test methods should be of the form:
        ``def test_method(self, graph, **kwargs)``
    '''

    #-------------------------------------------------------------------------#
    # Class properties

    @classmethod
    def setUpClass(cls):
        cls.graphs = []

    tolerance = 1e-5
    parser = XmlHandler()
    graphs = []
    instructions = []

    #-------------------------------------------------------------------------#
    # Instance

    def setUp(self):
        if not self.graphs:
            for graph_name in self.parser.get_graph_list(self.test_name):
                self.graphs.append(graph_name)

    @abstractproperty
    def test_name(self):
        pass

    def get_expected_result(self, graph, res_name):
        return self.parser.get_result(graph.name, res_name)

    @abstractmethod
    def gen_graph(self, graph_name):
        '''
        Must return a ``(g, di)`` tuple with `g` the graph instance and `di`
        the generation instructions (as a ``dict``) or ``None``.
        '''
        pass

